// Copyright (C) 2006-2013 Xilinx, Inc.  All rights reserved
// $Id: Data.c,v 1.75.206.8 2014/12/12 19:39:58 jcorbett Exp $
#include "Data.h"

#include <cstring> // for strncmp()
#include <algorithm>
#include <iomanip>
#include <sstream>

#include "XOS/Portability/export/Port_StringBase.h"
#include "XOS/Portability/export/Port_FileName.h"
#include "Databases/Rf/export/Facade/Rf_Node.h"

#include "main.h"
#include "ArgParser.h"
#include "DeviceModel.h"
#include "CadFile.h"
#include "Data.h"

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.75.206.8 $";

// Concatenate a set of strings into a single string with the given
// separator in between items.
const string ToString(const Classifier::NameSet& nameSet,
                      const string separator)
{
   ostringstream out;
   string sep;
   for (Classifier::NameSet::const_iterator nsIt(nameSet.begin());
        nsIt != nameSet.end();
        ++nsIt)
   {
      out << sep << *nsIt;
      sep = separator;
   }
   return out.str();
}

Data::~Data() {
   if (deviceModel) {
      delete deviceModel;
      deviceModel = 0;
   }
}

// Are we in UCF mode?
bool Data::NeedConstraintsCheck() const {
   return groupData.empty();
}

// Return a pointer to a DeviceModel instance, creating it if
// necessary.
DeviceModel* Data::GetDeviceModel() {
   if (!deviceModel) {
      if (NeedConstraintsCheck()) {
         // In UCF mode, load the device model for the given device
         // and package.
         deviceModel = new DeviceModel(device, package);
      } else {
         // In NCD mode, the device model is loaded as a side-effect
         // of loading the design. Extract it from the XCad_CadView.
         Data::GroupIt combined(groupData.find("combined"));
         ASSERT(combined != groupData.end());
         XCad_CadView* view(combined->second->GetView());
         deviceModel = new DeviceModel(view->GetDevice());
         combined->second->SetDeviceModel(deviceModel);
      }
      // During development at Xilinx, NCD files are produced that have no
      // package information associated with them. Attempting to access
      // package information of such a device file results in an
      // exception. This member function detects and reports this
      // condition.
      try {
         deviceModel->GetPackageColumnCount();
      } catch (...) {
         FatalInput(125, "Unable to obtain package information.\n"
                    "The input NCD file is not valid.");
      }
   }
   return deviceModel;
}

// Add one rectangular range to the site rectangle map.
void Data::AddRange(DeviceModel::SiteRect& siteRect) {
   string isolationGroup("<no isolation group>");
   ConstraintMap::iterator
      cmIt(constraintMap.find(siteRect.areaGroupName));
   if (cmIt != constraintMap.end()) {
      isolationGroup = cmIt->second;
      SiteRectMapIt srmIt(siteRectMap.find(isolationGroup));
      if (srmIt == siteRectMap.end()) {
         // Create a new range list and insert it into the map.
         DeviceModel::SiteRectVec rangeList;
         rangeList.push_back(siteRect);
         siteRectMap.insert(make_pair(isolationGroup, rangeList));
      } else {
         // Insert the range.
         srmIt->second.push_back(siteRect);
      }
   } else {
      const string& areaGroup(siteRect.areaGroupName);
      Warn(11, "No isolation group was found for area group "
             + areaGroup + ".");
   }
}

// Eliminate duplicates from the site rectangle map.
void Data::DeduplicateRanges() {
   for (SiteRectMapIt srmIt(siteRectMap.begin());
        srmIt != siteRectMap.end();
        ++srmIt)
   {
      DeviceModel::SiteRectVec& siteRange(srmIt->second);
      sort(siteRange.begin(), siteRange.end(), DeviceModel::SiteRectLess());
      DeviceModel::SiteRectVecIt
         endIt(unique(siteRange.begin(), siteRange.end(),
                      DeviceModel::SiteRectEqual()));
      siteRange.erase(endIt, siteRange.end());
   }
}

// Add an individual site to the site rectangle map.
void Data::AddSite(const string& areaGroupName, const string& siteName) {
   const Rf_Device* device(deviceModel->GetRfDevice());
   Rf_CellDef tileDef(device->GetTopCellDefTile());
   if (!tileDef) {
      FatalInternal(106, "Tile cell definition is not valid.");
   }
   Rf_CellDef tcd(device->GetTopCellDef());
   Rf_Cell site(deviceModel->GetSite(tcd, siteName));
   if (!site) {
      FatalInternal(117, "Unable to find tile (site) named " + siteName + ".");
   }

   DeviceModel::SiteRect siteRect;
   siteRect.areaGroupName = areaGroupName;
   siteRect.rangeString = siteName + ":" + siteName;

   AddRange(siteRect);
}

// Note: This function is needed but CollectAreaRangesFromUCF is not
// because ConstraintsChecker has access to the data instance, but the
// CadFile instance for the combined design does not.
void Data::CollectAreaRangesFromNCD() {
   for (GroupIt gIt(groupData.begin()); gIt != groupData.end(); ++gIt) {
      const IsolationGroupName& group(gIt->first);
      if (group != "combined") {
         siteRectMap.insert(make_pair(group, gIt->second->GetSiteRectList()));
      }
   }
}

// Get the file name associated with a net. This is O(n) where n
// is the total number of nets in the given group, that is to say,
// slow, but only called for diagnostics on bad user input.
const string Data::GetFileName(XCad_CadNetwork* net,
                               IsolationGroupName& group)
{
   pair<GroupIt, GroupIt> range(groupData.equal_range(group));
   for (GroupIt groupIt(range.first);
        groupIt != range.second;
        ++groupIt)
   {
      CadFile::NetVec& nets(groupIt->second->GetNets());
      if (find_if(nets.begin(), nets.end(), NetMatch(net)) != nets.end()) {
         return groupIt->second->GetFileName();
      }
   }
   const string netName(net->GetReportableName());
   FatalInternal(41, "No file name found for net " + netName + ".");
   return "<missing group " + group + " file name>";
}

// The previous group for this user net is defined but it is not
// "ignored" or "combined" thus it must be a user group.
void Data::AddGroupConflict(XCad_CadNetwork* net,
                            const string& group,
                            const string& fileName)
{
   NetGroupIt prevGroupIt(netGroup.find(net));
   // Record everything here because the group will be changed to "ignored".
   GroupConflict gc;
   gc.net1 = net;
   gc.group1 = group;
   gc.fileName1 = Port::FileName(fileName.c_str()).GetTail();
   gc.group2 = "<no group>";
   gc.fileName2 = "<no file>";
   if (prevGroupIt != netGroup.end()) {
      gc.net2 = prevGroupIt->first;
      gc.group2 = prevGroupIt->second;
      gc.fileName2 =
         Port::FileName(GetFileName(prevGroupIt->first,
                                    prevGroupIt->second).c_str())
         .GetTail();
   }
   groupConflictList.push_back(gc);
}

// XCad_CadNetwork::GetFilteredPins requires a predicate. This functor
// filters pins according to the result of XCad_CadPin::IsSignalDriver().
struct SourcePinFilter {
   bool operator()(const XCad_CadPin* pin) {
      return pin->IsSignalDriver();
   }
};

// XCad_CadNetwork::GetFilteredPins requires a predicate. This functor
// filters pins according to the result of XCad_CadPin::IsSignalDriver().
struct LoadPinFilter {
   bool operator()(const XCad_CadPin* pin) {
      return !pin->IsSignalDriver();
   }
};

// Get the isolation group containing block unless block is not
// isolated, in which case return the empty string.
Data::IsolationGroupName Data::GetIsolationGroup(const XCad_CadBlock* block) {
   IsolationGroupName result("");
   const XCad_CadBlock* parent(block);
   while (parent && result.empty()) {
      for (GroupIt gIt(groupData.begin()); gIt != groupData.end(); ++gIt) {
         const IsolationGroupName& group(gIt->first);
         if (group != "combined") {
            if (gIt->second->GetModule() == parent) {
               result = group;
            }
         }
      }
      parent = parent->GetParent();
   }
   return result;
}

typedef std::vector<const XCad_CadPin*> CadPins;

bool IsPhysicalPin(const XCad_CadPin* pin) { return pin->IsRoutable(); }

// R#check_irs_loads - Return a string listing the source and loads for a net.
// 1. Find the pins for which IsRoutable() returns true. These pins
//    correspond to actual hardware, not design hierarchy.
// 2. Determine whether the pin is a driver or a load.
// 3. Follow the parent chain until reaching an isolation group or
//    top.
// 4. Report the group. The net is reported by the caller.
const string Data::DescribeNetConnections(XCad_CadNetwork* net) {
   string result;
   string prev("");
   bool multipleLoads(false);
   CadPins pins;
   net->GetFilteredPins(pins, IsPhysicalPin);
   sort(pins.begin(), pins.end());
   CadPins::const_iterator end(unique(pins.begin(), pins.end()));
   Classifier::NameSet loads;
   for (CadPins::const_iterator pinIt(pins.begin());
        pinIt != end;
        ++pinIt)
   {
      const XCad_CadPin* pin(*pinIt);
      const XCad_CadBlock* block(pin->GetParent());
      string group(GetIsolationGroup(block));
      if (group.empty()) {
         group = "Top-level";
      }
      if (pin->IsSignalLoad()) {
         if (prev.empty()) {
            prev = group;
         } else if (prev != group) {
            multipleLoads = true;
         }
      }
      const string direction(pin->IsSignalLoad() ? "Load " : "Source ");
      string report("    " + direction + "group: " + group);
      if (pin->IsSignalLoad()) {
         loads.insert(report);
      } else {
         result += report + "\n";
      }
   }
   result += ToString(loads, "\n") + "\n";
   if (multipleLoads) {
      ++irsMultiLoadCount;
      Report("Isolation Violation:128: Inter-region signal "
             + string(net->GetUserHierName())
             + "\nhas loads in more than one isolation group.");
   }

   return result;
}

void Data::ReportNetConnections(XCad_CadNetwork* net) {
   Report(DescribeNetConnections(net));
}

// R#report_interregion_signals
void Data::ReportUserGlobalNets() {

   Section("Uncategorized User Global Nets");

   Report("  The nets below are found in multiple isolation groups,\n"
          "  therefore it is incumbent upon the user to prove these\n"
          "  signals do not violate data isolation requirements. Only\n"
          "  power, ground, global clocks, trusted inter-region signals,\n"
          "  or explicitly permitted control signals may be global.\n");

   sort(groupConflictList.begin(), groupConflictList.end(),
        GroupConflictLess());
   NetSet topLevelNets;
   for (GroupConflictListIt gcIt(groupConflictList.begin());
        gcIt != groupConflictList.end();
        ++gcIt)
   {
      topLevelNets.insert(gcIt->net1);
   }

   Report("  Total uncategorized user global nets: " +
          ToString(topLevelNets.size()) + "\n");

   for (GroupConflictListIt gcIt(groupConflictList.begin());
        gcIt != groupConflictList.end();
        ++gcIt)
   {
      Report("  " + string(gcIt->net1->GetReportableName()));
      ReportNetConnections(gcIt->net1);
   }
   if (groupConflictList.empty()) {
      Report("  No uncategorized nets were found in multiple "
             "isolation groups.");
   }
}

// R#report_range_constraints - Report area range constraints found in
// the ncd file(s).
void Data::ReportSiteRanges() {

   Section("Area Range Constraints");

   const string leftMargin(2, ' ');
   const string colsep(4, ' ');
   size_t igWidth(16);
   size_t agWidth(20);

   Report(
"   Isolation Group              Area Group    Ranges\n"
"  ----------------    --------------------    ---------------------------");
   for (SiteRectMapIt srmIt(siteRectMap.begin());
        srmIt != siteRectMap.end();
        ++srmIt)
   {
      IsolationGroupName group(srmIt->first);
      DeviceModel::SiteRectVec rangeList(srmIt->second);
      ostringstream out;
      out << leftMargin << setfill(' ') << setw(igWidth) << group;
      if (igWidth < group.size()) {
         out << '\n' << leftMargin << string(igWidth, ' ');
      }
      string sep(colsep);
      string prevAreaGroupName;
      // List the area groups and ranges. Only list the area group
      // name once at the beginning of a series of lines of area ranges.
      for (DeviceModel::SiteRectVecIt rIt(rangeList.begin());
           rIt != rangeList.end();
           ++rIt)
      {
         string ag(prevAreaGroupName == rIt->areaGroupName ?
                   "" : rIt->areaGroupName);
         out << sep << setw(agWidth) << ag << colsep
             << rIt->rangeString;
         sep = ",\n" + leftMargin + string(igWidth, ' ') + colsep;
         prevAreaGroupName = rIt->areaGroupName;
      }
      Report(out.str());
   }
   if (siteRectMap.empty()) {
      Report("  No area groups found.");
   }
}

// R#report_interregion_signals - Report the 'ignored' nets,
// categorized by reason for ignoring them.
//
// "Ignored/unused nets" is not really an ideal description.
// Originally it meant nets that did not need to be
// searched during cost-based analysis.  Now it means, highlighted as
// global or confined to a single tile, but the nets confined to a
// single tile are not reported at all, so these are just global nets
// and in tile-based analysis, all nets in the fence must be examined
// by the evaluators. The contents of each non-empty fence tile must
// be considered since there is no automated way of distinguishing
// acceptable interregion signals from unacceptable ones.
void Data::ReportUnusedNets() {
   Section("Categorized Nets");

   Report("  The nets listed below present lesser risk than uncategorized\n"
          "  nets due to their physical extents or signal sources.");

   for (WhyIgnoreIt ignoreIt(whyIgnoreMap.begin());
        ignoreIt != whyIgnoreMap.end();
        ++ignoreIt)
   {
      const string& reason(ignoreIt->first);
      if (reason != classifier.GetHeading(Classifier::InsideBusMacro) &&
          reason != classifier.GetHeading(Classifier::InterRegionSignal) &&
          reason != classifier.GetHeading(Classifier::DedicatedRoute))
      {

         Subsection(reason);

         ReportSet(ignoreIt->second,
                   "  No nets were categorized for this reason.");
      }
   }
}

// When NCD files are merged, occasionally net
// names will be discarded. This is expected to happen
// for bus macro inputs and clock nets.
void Data::AddDiscardedNet(XCad_CadNetwork* net,
                           XCad_RMed* rmed,
                           const std::string& fileName)
{
   Classifier::NetCode reasonCode(classifier.Classify(net, rmed));
   string heading(classifier.GetHeading(reasonCode));
   if (reasonCode == Classifier::NoClassification) {
      heading = "Nets";
   } else if (reasonCode == Classifier::SourceRegionalClock) {
      heading += " in File " + fileName;
   } else {
      string::size_type pos(heading.find('\n'));
      if (pos != string::npos) {
         heading = heading.substr(0, pos);
      }
   }
   heading += " in File " + fileName + "\n\n"
      "  The following nets were not found in the combined\n"
      "  design and thus do not contribute to the final "
      "bitstream.\n"
      "  Either they were trimmed during the merge process or a\n"
      "  group NCD from an unrelated design was supplied.";

   whyIgnoreMap[heading].insert(string(net->GetReportableName()));
}

// R#report_interregion_signals - For every net in the given group,
// create a mapping from it to the group. Maintain the count of
// "combined" entries for error checking.
void Data::LabelNetGroups(const string group, int& combinedCount) {
   // The group parameter comes from groupData keys. The keys
   // are "combined" and the -group parameters. There's an explicit
   // check in class ArgParser that the -group switch not have
   // the value "ignored".
   ASSERT(group != "ignored");

   GroupIt groupIt(groupData.find(group));
   if (groupIt == groupData.end()) {
      FatalInternal(81, "Data from group " + group + " was not found.");
   }
   string fileName(groupIt->second->GetFileName());
   CadFile::NetVec& nets(groupIt->second->GetNets());
   XCad_RMed* rmed(groupIt->second->GetRMed());
   for (CadFile::NetVecIt netIt(nets.begin()); netIt != nets.end(); ++netIt) {
      XCad_CadNetwork* net(*netIt);
      NetGroupIt prevGroupIt(netGroup.find(net));
      if (group == "combined") { // First encounter with net.
         ASSERT(prevGroupIt == netGroup.end());
         ++combinedCount;
         netGroup.insert(make_pair(net, "combined"));
      } else {                  // On to the user's groups...
         if (prevGroupIt == netGroup.end()) {
            AddDiscardedNet(net, rmed, fileName);
         } else if (prevGroupIt->second == "ignored") {
            ;                       // do nothing
         } else if (prevGroupIt->second == "combined") {
            // Trust the design software to give us each net only once.
            --combinedCount;
            Classifier::NetCode reasonCode(classifier.Classify(net, rmed));
            if (reasonCode == Classifier::NoClassification) {
               prevGroupIt->second = group; // Net isolated to this group.
            } else {
               // Record the categorization of this net.
               const string heading(classifier.GetHeading(reasonCode));
               whyIgnoreMap[heading].insert(string(net->GetReportableName()));
               prevGroupIt->second = "ignored";
            }
         } else if (prevGroupIt->second == group) {
            // Shouldn't get any net more than once in a group, but let's
            // hear about it if we do.
            const string& netName(net->GetReportableName());
            Warn(78, "Net " + netName
                 + " appears more than once in group "
                 + group + ".\nCheck the command line arguments for"
                 " duplicated file names.");
         } else {
            // Previous group is not "combined" and not "ignored";
            // current group is not "combined" and not "ignored".
            // Definitely global. Possibly a conflict....
            Classifier::NetCode reasonCode(classifier.Classify(net, rmed));
            string heading(classifier.GetHeading(reasonCode));
            if (reasonCode == Classifier::NoClassification) {
               // No reason for this net to appear in multiple groups.
               AddGroupConflict(net, group, fileName);
            } else {
               heading = "Global " + heading;
               whyIgnoreMap[heading].insert(string(net->GetReportableName()));
            }
            // Mark it ignored to avoid redundant error messages.
            prevGroupIt->second = "ignored";
         }
      }
   }
}

// Write a section to the output report that consists of a heading and
// a list of items or a message indicating there are no items under
// this heading.
void Data::ReportSet(const Classifier::NameSet& nameSet,
                     const string emptyMessage)
{
   for (Classifier::NameSetIt setIt(nameSet.begin());
        setIt != nameSet.end();
        ++setIt)
   {
      Report("  " + *setIt);
   }
   if (nameSet.empty()) {
      Report(emptyMessage);
   }
}

// If there are left over "combined" nets after inserting the other
// nets, report the combined nets as ambiguous.
//
// Automatically inserted DCM logic sometimes appears in the combined
// design but not the PRMs. All other nets in the combined design
// should be in at least one of the PRMs.
void Data::CheckCombinedCount(int combinedCount) {
   if (combinedCount != 0) {
      // Loop over the all the nets in the combined group and
      // collect the ones that have not had their group overwritten.
      GroupIt groupIt(groupData.find("combined"));
      if (groupIt == groupData.end()) { // R#report_errors
         FatalInternal(20, "Data from combined group was not found.");
      }
      CadFile::NetVec& netVec(groupIt->second->GetNets());
      XCad_RMed* rmed(groupIt->second->GetRMed());
      for (CadFile::NetVecIt netVecIt(netVec.begin());
           netVecIt != netVec.end();
           ++netVecIt)
      {
         XCad_CadNetwork* net(*netVecIt);
         NetGroupIt missingIt(netGroup.find(net));
         if (missingIt == netGroup.end()) { // R#report_errors
            const string netName(net->GetReportableName());
            FatalInternal(40, "Found net " + netName + " with no group.");
         }
         if (missingIt->second == "combined") {
            // R#report_interregion_signals
            string heading(classifier.GetHeading(net, rmed));
            const string netName(net->GetReportableName());
            if (heading.empty()) {
               FatalInternal(130, "Unable to categorize net " + netName
                             + ".\n"
                             + DescribeNetConnections(net));
            }
            whyIgnoreMap[heading].insert(netName);
         }
      }
   }
}

// R#report_interregion_signals - Collect pointers to all the bus
// macro blocks. Within each bus macro block there are nets with
// the property IS_BUS_MACRO. This property is only used for such
// nets.
void Data::CollectBusMacros() {
   GroupIt groupIt(groupData.find("combined"));
   if (groupIt == groupData.end()) {
      FatalInternal(104, "Data from group 'combined' was not found.");
   }
   classifier.CollectBusMacros(groupIt->second->GetNets());
}

// R#report_interregion_signals - Report the bus macro usage. List the
// bus macro instance names and source macro name. Report each pin and
// the net connected to it, if any. Also list the groups of all
// block the net is connected to, except the bus macro block.
void Data::ReportBusMacroUsage() {

   Section("Trusted Bus Macros");

   Report(
      "  The isolation groups listed for each Trusted Bus Macro pin are\n"
      "  are determined by the blocks the net is connected to.\n"
      "  Therefore a pin may be associated with more than one group if\n"
      "  the net connected to it is connected to blocks in more than\n"
      "  one group.\n");
   for (Classifier::BlockSetIt bsIt(classifier.BusSetBegin());
        classifier.BusSetItValid(bsIt);
        ++bsIt)
   {
      XCad_CadBlock* macroBlock(*bsIt);
      XCad_CadBlock* topBlock(macroBlock->GetParent());
      ASSERT(topBlock);
      string source(classifier.GetBusMacroSource(macroBlock));
      Classifier::BlockName busName(macroBlock->GetName());
      Report("\n"
             "  Trusted Bus Macro " + busName);
      Report("    Source: " + source);
      Report("\n"
             "    Pin           Isolation Group(s)  Net\n"
             "    ------------  ------------------  ------------------------");
      Classifier::NameSet pinSet;

      // List each pin of the bus macro, the net attached to the
      // pin, and the isolation group of the net. Nets that
      // are not isolated are listed as "combined" (aka in the top
      // level design).
      const string topLevel("Top-level");
      typedef XCad_CadBlock::Pins PinVec;
      PinVec pins(macroBlock->GetPins());
      for (PinVec::iterator pinIt(pins.begin()); pinIt != pins.end(); ++pinIt)
      {
         XCad_CadPin* pin(*pinIt);
         XCad_CadSignal* signal(pin->GetSignal());
         ostringstream out;
         out << "    " << setw(12) << left << string(pin->GetName()) << "  ";
         if (signal) {
            XCad_CadNetwork* net(signal->GetNetwork());
            ASSERT(net);
            Classifier::NameSet groupSet;
            if (classifier.IsDrivenByGlobalClock(net)) {
               // Nets driven by global clocks are always at the top level.
               groupSet.insert(topLevel);
            } else {
               XCad_CadNetwork::Signals sigs(net->GetSignals());
               for (XCad_CadNetwork::Signals::iterator sigIt(sigs.begin());
                    sigIt != sigs.end();
                    ++sigIt)
               {
                  XCad_CadSignal::Pins pins((*sigIt)->GetPins());
                  for (XCad_CadSignal::Pins::iterator pinIt(pins.begin());
                       pinIt != pins.end();
                       ++pinIt)
                  {
                     XCad_CadPin* pin(*pinIt);
                     XCad_CadBlock* block(pin->GetParent());
                     if (block != topBlock) {
                        // Search groupData for a CadFile with module
                        // matching some parent of this pin.
                        XCad_CadBlock* p(block);
                        while (p && p != macroBlock) {
                           p = p->GetParent();
                        }
                        if (p != macroBlock) {
                           // Do we need to search the parents of
                           // block != macroBlock as well?
                           IsolationGroupName group(GetIsolationGroup(block));
                           if (group.empty()) {
                              group = topLevel;
                           }
                           groupSet.insert(group);
                        } else {
                           ;    // This is the pin connected to the bus macro.
                        }
                     } else {
                        // The net is not from an isolated module. If
                        // this case ever occurs, it could be a
                        // connection to a global I/O pin.
                        groupSet.insert(topLevel);
                     }
                  }
               }
            }
            string groupText(ToString(groupSet));
            if (groupSet.empty()) {
               groupText = "No loads";
            }
            out << setw(20) << groupText + "  ";
            Classifier::NetName netName(net->GetReportableName());
            out << netName;
         } else {
            out << setw(20) << "-";
            out << "-";
         }
         pinSet.insert(out.str());
      }
      Report(ToString(pinSet, "\n"));
   }
   if (!classifier.BusSetItValid(classifier.BusSetBegin())) {
      Report("  No trusted bus macros were found.");
   }
}

// R#report_interregion_signals - Label nets with their corresponding
// group. Label the "combined" nets first because they define the
// universe of discourse, i.e. every other group must be a subset of
// the "combined" group.
//
// Also, classify (categorize) nets according to their usage in
// the design, e.g. constant, attached to clocks, attached to bus
// macros, etc.
void Data::LabelNets() {
   int combinedCount = 0;
   LabelNetGroups("combined", combinedCount);
   for (GroupIt groupIt(groupData.begin());
        groupIt != groupData.end();
        ++groupIt)
   {
      if (groupIt->first == "combined") {
         // Previously the condition also tested groupIt->first == "ignored"
         // but that condition would never be true. See ArgParser.c where
         // CadFile instances are created.
         continue;
      }
      LabelNetGroups(groupIt->first, combinedCount);
   }
   CheckCombinedCount(combinedCount);
}

int Data::GetErrorCount() {
   return unroutedCount + omittedPS7 + irsMultiLoadCount;
}

// Return 0 for success and 1 for failure.
int Data::GetStatus() {
   return !!GetErrorCount();
}
