// Copyright (C) 2009-2013 Xilinx, Inc.  All rights reserved
// $Id: TileChecker.c,v 1.42.206.36 2013/12/10 19:57:38 jcorbett Exp $
#include <algorithm>
#include <iomanip>
#include <sstream>
#include <limits>

#include "Databases/Rf/export/Facade/Rf_Device.h"
#include "Databases/Rf/export/Facade/Rf_Cell.h"
#include "Databases/Rf/export/Facade/Rf_CellContainer.h"
#include "XCad/export/XCad_RMed.h"
#include "XCad/export/RMed/XCad_RMedRug.h"
#include "GenericAlgorithms/Utilities/export/UtilCoord.h"
#include "TUtilities/export/regex.hpp"

#include "ArgParser.h"
#include "CadFile.h"
#include "Svg.h"
#include "Edge.h"
#include "ContourMap.h"
#include "main.h"
#include "Classifier.h"
#include "TileChecker.h"

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.42.206.36 $";

// R#check_irs_separation, R#check_interconnect_routing,
// R#check_direct_routing, R#check_logic_tiles
TileChecker::TileChecker(Data& data)
   : data(data),
     deviceModel(data.GetDeviceModel()),
     diagramEmitter(data,
                    contourMap,
                    grid,
                    groupSetFamily,
                    superIntSet,
                    superBoundaryMap,
                    superGlobalMap,
                    configCenterGroupSet),
     netAdjacencyFaultCount(0),
     areaRangeFaultCount(0),
     logicAdjacencyFaultCount(0),
     netContentFaultCount(0),
     logicContentFaultCount(0),
     configCenterFaultCount(0),
     dspFaultCount(0),
     impliedSiteCount(0),
     topLevelCmtCount(0),
     superFaultCount(0),
     gtxFaultCount(0),
     irsNetPipCount(0),
     irsMultiLoadCount(0),
     clockFaultCount(0),
     configColumnFaultCount(0),
     configCenterGroupSet(0),
     configCenterNeighborGroupSet(0),
     nominalFenceNets(0),
     sweptConfigCenterColumn(false),
     sweptClockBufferColumn(false),
     configCenterRowMin(static_cast<size_t>(-1)),
     configCenterRowMax(0)
{
   // The Rf_Cell class represents configuration information at a
   // specific location in the device. The configuration information
   // may be a mode of a logic element or a node of routing. (A node
   // is a unit of wiring with no internal switches. It may span
   // multiple tiles.)
   //
   // The cells are organized in a fixed hierarchy: device level, tile
   // level, site level, and leaf level. This class is concerned with
   // the tile level, but needs to translate the site level
   // information from user constraint file (UCF) to tiles.

   // This allows space for all the tile level cells, even though we
   // are primarily interested in the interconnect tiles.
   Port::UINT32 columns(deviceModel->GetTileColumnCount());
   Port::UINT32 rows(deviceModel->GetTileRowCount());
   grid.resize(columns * rows, 0);
   contourMap.reset(new ContourMap<Tile>(columns, rows, grid));

   // Fill our grid with all the cells, so we can access them based on
   // their geometric arrangement.
   Rf_CellContainer cells(deviceModel->GetCells());
   for (Rf_CellContainer::const_iterator i(cells.begin());
        i != cells.end();
        ++i)
   {
      Rf_Cell cell(*i);
      ASSERT(cell);
      Port::UINT32 x(cell.GetTileColumn());
      Port::UINT32 y(cell.GetTileRow());
      Tile* tile(new Tile(cell,
                          netAdjacencyFaults,
                          sweepState,
                          groupSetFamily,
                          data.NeedConstraintsCheck()));
      contourMap->MarkPoint(x, y, tile);
   }

   // Display interconnect tiles that can be used as fence tiles and
   // supersites that occupy more than one tile.
   MapHiddenStripes();
   AllocateSingletonGroups();
}

// Delete each of the Tile instances created in the constructor above.
TileChecker::~TileChecker() {
   for (TileVecIt gIt(grid.begin()); gIt != grid.end(); ++gIt) {
      Tile* tile(*gIt);
      // It's theoretically possible that not every position in the
      // grid would have a pointer to a Tile instance. In practice,
      // the grid has been fully populated for Virtex-4, Virtex-5, and
      // Spartan-6.
      if (tile) {
         delete *gIt;
      }
   }
}

// R#check_irs_separation, R#check_interconnect_routing - The
// isolation analysis rests on the existence of empty interconnect
// tiles between isolated regions. Return true for any tile that is
// not an interconnect tile and false otherwise. This function is used
// to filter the tile data to produce a floor plan of the interconnect
// tiles.
//
// The "tiles" in the IVT floorplan diagrams are an abstraction of
// several underlying hardware blocks: interconnect, interface, and
// logic. Tiles exclusively dedicated to global clock routing or
// configuration routing are not shown.
bool TileChecker::IsHiddenTile(size_t x, size_t y) {
   Tile* tile(contourMap->At(x, y));
   bool interconnect(tile &&
                     deviceModel->IsInterconnectCell(tile->GetCell()));
   return !interconnect;
}

// R#check_irs_separation, R#check_interconnect_routing, R#draw_fence -
// Sweep through the grid looking for rows and columns that are
// entirely composed of hidden tiles. Mark these in the contour map.
void TileChecker::MapHiddenStripes() {
   for (size_t y(0); y < contourMap->RowCount(); ++y) {
      bool isHiddenRow(true);
      for (size_t x(0);
           x < contourMap->ColumnCount();
           ++x)
      {
         if (!IsHiddenTile(x, y)) {
            isHiddenRow = false;
            break;
         }
      }
      if (isHiddenRow) {
         contourMap->HideRow(y);
      }
   }
   size_t configCenterX(static_cast<size_t>(-1));
   size_t configCenterCount(0);
   if (1.0 < diagramEmitter.GetSmallZynqOffset()) {
      // The config center area is 7 tiles wide.  For devices in which
      // the config center area consumes the entire height of the
      // device, we must expose six columns that would otherwise be
      // hidden because they do not contain any visible tiles. The
      // config center column itself is already counted.
      configCenterCount = 6; 

      for (size_t y(0); y < contourMap->RowCount(); ++y) {
         for (size_t x(0);
              x < contourMap->ColumnCount();
              ++x)
         {
            Tile* tile(contourMap->At(x, y));
            if (tile) {
               const string tileType(tile->GetCell().GetCellDef().GetName());
               if (tileType == "CFG_CENTER_TOP") {
                  configCenterX = x - configCenterCount;
                  goto gotX;
               }
            }
         }
      }
   }
  gotX:
   for (size_t x(0); x < contourMap->ColumnCount(); ++x) {
      bool isHiddenColumn(true);
      if (configCenterX <= x && configCenterCount) {
         isHiddenColumn = false;
         --configCenterCount;
      } else {
         for (size_t y(0); y < contourMap->RowCount(); ++y) {
            if (!IsHiddenTile(x, y)) {
               isHiddenColumn = false;
               break;
            }
         }
      }
      if (isHiddenColumn) {
         contourMap->HideColumn(x);
      }
   }

   contourMap->AdjustGrid();
}

// Don't rely on luck to get groupSetFamily populated.  If the first
// time a group is seen is after the groupSetList is sorted, we have a
// problem. This rarely happens, but it is possible, for example, if
// the first time a group is processed it is in the context of a
// violation.
void TileChecker::AllocateSingletonGroups() {
   for (Data::GroupIt groupIt(data.groupData.begin());
        groupIt != data.groupData.end();
        ++groupIt)
   {
      const string group(groupIt->first);
      if (group != "combined") {
         // The result will be cached; we don't need the return value.
         groupSetFamily.FindOrMakeSet(group);
      }
   }
}

// R#check_irs_separation, R#check_interconnect_routing - The Rf_Wire
// class represents a portion of a node that occupies exactly one
// tile.
//
// Map wire and net information to a tile. Update the set of
// isolation groups that occupy the tile based on the block hierarchy
// of the routing, not the area range constraints. Mark global signals
// in the tile.
void TileChecker::MapOneWireTile(Rf_Node rfNode,
                                 Rf_Wire wire,
                                 const string& netName,
                                 const string& netGroup,
                                 XCad_CadNetwork* net)
{
   Rf_Cell cell(wire.GetCell());
   if (!cell) {
      FatalInternal(99, "Device model problem: undefined cell.");
   }

   string wireName(wire.GetName());
   Tile* tile(diagramEmitter.GetTile(cell));

   // Generate a label for the node that helps the user locate it
   // in FPGA Editor.
   const string nodeName(rfNode.GetName());
   const string nodeId(ToString(rfNode.GetIndex()));
   bool isGlobal(netGroup == "combined" || netGroup == "ignored");
   const string scopeLabel(isGlobal
                           ? string("Top-level Net: ")
                           : "Group " + netGroup + " Net: ");
   const string netLabel(scopeLabel + netName
                         + "\n      Node: " + nodeName + ":" + nodeId
                         + " Wire: " + wireName);
   tile->AddNetDescription(netLabel);

   if (isGlobal) {
      tile->SetHasGlobalNet();
      tile->AddNet(net);

      // Check for PIPs in global nets.
      //
      // Recall a PIP is a programmable interconnect
      // point. GetArcCount() returns the number of potential
      // connections this wire has to other nodes in this tile. (An
      // arc is simply a pair of nodes with a programmable switch
      // controlled by a configuration memory element connecting
      // them.)
      if (0 < wire.GetArcCount() &&
          deviceModel->IsInterconnectCell(cell) &&
          !data.classifier.IsDrivenByGlobalClock(net) &&
          !data.classifier.IsDrivenByRegionalClockBuffer(net))
      {

         // A global route composed of more than one node is
         // only problematic if a tile containing the interconnect
         // point between nodes of the route is needed as a fence
         // tile. The used PIP makes it invalid as a fence tile.
         tile->AddTerminal(net, wire);
      }
   } else {
      // Isolated.
      tile->SetHasIsolatedNet();

      // Add the given group to the set of groups represented in this
      // tile, if it is not already present.
      Data::GroupSet* gs0(tile->GetGroupSet());
      Data::GroupSet* gs1(groupSetFamily.FindOrMakeSet(netGroup, gs0));
      tile->SetGroupSet(gs1);
   }
   if (data.classifier.IsInsideBusMacro(net)) {
      tile->SetHasBusMacro();
   }
}

// R#check_irs_separation, R#check_interconnect_routing - Map one
// isolation group to tiles. This routine iterates over all the
// nets in the given group and collects net information per
// tile.
void TileChecker::MapIsolationGroupToTiles(const string& group) {
   ASSERT(group != "ignored");
   CadFile* cadFile(data.groupData.find(group)->second.get());
   // For each net in the group...
   CadFile::NetVec& nets(cadFile->GetNets());
   // RMed stands for resource mediator.
   XCad_RMed* rmed(cadFile->GetRMed());
   for (CadFile::NetVecIt netIt(nets.begin());
        netIt != nets.end();
        ++netIt)
   {
      XCad_CadNetwork* net(*netIt);
      const string netName(net->GetUserHierName());
      const string netGroup(data.netGroup.find(net)->second);
      // Rug stands for resource usage graph.
      XCad_RMedRug* rug(rmed->GetRug(net));
      if (rug) {
         // For each node in the net...
         const XCad_RMedRugNode* next;
         for (const XCad_RMedRugNode* node(rug->GetFirstRugNode());
              node != NULL;
              node = next)
         {
            next = node->GetNextRugNode();
            Rf_Node rfNode(node->GetRfNode());
            if (rfNode) {
               // For each wire (or tile) of the node,
               // mark the tile based on the properties of the net.
               Rf_WireContainer wires(rfNode.GetWires());
               for (Rf_WireContainer::const_iterator wIt(wires.begin());
                    wIt != wires.end();
                    ++wIt)
               {
                  Rf_Wire wire(*wIt);
                  ASSERT(wire); // Very unlikely we'll be handed a bad wire.
                  MapOneWireTile(rfNode,
                                 wire,
                                 netName,
                                 netGroup,
                                 net);
               }
            } else {
               ;
               // There are some nodes named FAKE_RFNODE associated
               // with GLOBAL_LOGIC0 and GLOBAL_LOGIC1.
            }
         }
      }
   }
}

// R#check_irs_separation, R#draw_fence, R#check_interconnect_routing -
// In Xilinx parlance, a net is a set of routing resources all
// of which a programmed to be electrically connected together. A
// net is comprised of a set of nodes. Electrical connections are
// made between pairs of nodes (called arcs) that are typically
// implemented as a single pass transistor controlled by a
// configuration memory cell. Nodes are further decomposed into wires
// according to tile boundaries, with each wire occupying exactly one
// tile. A tile is an ASIC cell that is associated with a rectangular
// region used to represent the device in graphical form.
//
// This function decomposes nets into their constituent wires and
// thus maps them into all the tiles they occupy. The "combined" group
// is processed first so that top-level signals (nets that are not
// isolated to a single region) do not generate false isolation
// conflicts. Note however that these top-level nets are marked as
// such.
//
// It is assumed that the designers and evaluators will examine the
// list of tiles in the fence to insure that only signals allowed to
// span the fence in fact do so.
//
// It is expected that a tile may contain
//    - no signals,
//    - signals from a single isolation group,
//    - global signals, or
//    - signals from a single isolation group and global signals.
// It is not expected that a tile will contain signals from more than
// one isolation group. This is an isolation violation.
void TileChecker::MapNetsToTiles() {
   MapIsolationGroupToTiles("combined");
   // For each isolation group...
   for (Data::GroupIt groupIt(data.groupData.begin());
        groupIt != data.groupData.end();
        ++groupIt)
   {
      ConsoleDot();
      const string group(groupIt->first);
      if (group == "combined") {
         continue;
      }
      MapIsolationGroupToTiles(group);
   }
}

// Add a neighbor to a supersite for later adjacency testing.
// Do this asymmetrically so we can safely assume that a site is a
// supersite if it has super neighbors.
void TileChecker::AddNeighbor(size_t tx, size_t ty, Tile* superSite) {
   if (contourMap->InGrid(tx, ty)) {
      if (!IsHiddenTile(tx, ty)) {
         Tile* neighbor(contourMap->At(tx, ty));
         superSite->AddNeighbor(neighbor);
      }
   }
}

// Take one step along the perimeter of the site that contains grid.at(i),
// heading in the given direction.
//
// Invariants
//    1. grid.at(iPrev) is inside the perimeter.
//    2. Looking from previous to current, the inside is on the right.
//
// At most three successor tiles must be examined to find one that is
// just inside the perimeter, while its left neighbor (relative to
// the direction of travel) is outside. To do this, successors are
// considered in order:
//    1. Forward and left
//    2. Straight forward
//    3. Forward and right
//
// We follow the inside of the perimeter because there are always
// cells inside the perimeter, but not always outside (GTPs are at the
// edges of the die).
//
// Note: Since we have convenient access to the neighbor tiles outside
// the perimeter, we add them to a set for later isolation analysis.
void TileChecker::StepAlongPerimeter(Tile* superSite,
                                     size_t iPrev,
                                     Direction dirPrev,
                                     size_t& iNext,
                                     Direction& dirNext)
{
   Rf_Cell cell(grid.at(iPrev)->GetCell());
   unsigned int itx(cell.GetTileColumn()), ity(cell.GetTileRow());
   ASSERT(superSiteSet.count(make_pair(itx, ity)));

   int delta0, delta1;
   Direction nextDir0, nextDir1;
   switch (dirPrev) {
   case Left: case Right:
   {
      if (dirPrev == Left) {
         //
         //     +---+
         //     | i |   inside, iPrev
         //    <-----<  perimeter
         //     | o |   outside
         //     +---+
         //
         delta0 = 1; delta1 = -1; nextDir0 = Down; nextDir1 = Up;
      } else {
         //
         //     +---+
         //     | o |
         //    >----->
         //     | i |
         //     +---+
         //
         delta0 = -1; delta1 = 1; nextDir0 = Up; nextDir1 = Down;
      }
      unsigned int otx(itx), oty(contourMap->NextVisibleRow(itx, ity, delta0));
      ASSERT(!superSiteSet.count(make_pair(otx, oty)));
      AddNeighbor(otx, oty, superSite);
      unsigned int ltx(contourMap->NextVisibleColumn(otx, oty, delta1));
      unsigned int lty(oty);
      if (superSiteSet.count(make_pair(ltx, lty))) {
         dirNext = nextDir0;
         iNext = contourMap->Index(ltx, lty);
      } else {
         AddNeighbor(ltx, lty, superSite);
         unsigned int stx(contourMap->NextVisibleColumn(itx, ity, delta1));
         unsigned int sty(ity);
         if (superSiteSet.count(make_pair(stx, sty))) {
            dirNext = dirPrev;
            iNext = contourMap->Index(stx, sty);
         } else {
            AddNeighbor(stx, sty, superSite);
            dirNext = nextDir1;
            iNext = iPrev;
         }
      }
      break;
   }
   // The Up/Down case is only different from the Left/Right case in
   // that row and column operations are exchanged.
   case Up: case Down:
   {
      if (dirPrev == Up) {
         //         ^
         //     +---|---+
         //     | o | i |
         //     +---|---+
         //         ^
         delta0 = -1; delta1 = -1; nextDir0 = Left; nextDir1 = Right;
      } else {
         //         V
         //     +---|---+
         //     | i | o |
         //     +---|---+
         //         V
         delta0 = 1; delta1 = 1; nextDir0 = Right; nextDir1 = Left;
      }
      // Note: this code matches the Left/Right code above, but with
      // NextVisibleRow() swapped with NextVisibleColumn().
      unsigned int otx(contourMap->NextVisibleColumn(itx, ity, delta0));
      unsigned int oty(ity);
      ASSERT(!superSiteSet.count(make_pair(otx, oty)));
      AddNeighbor(otx, oty, superSite);
      unsigned int ltx(otx);
      unsigned int lty(contourMap->NextVisibleRow(otx, oty, delta1));
      if (superSiteSet.count(make_pair(ltx, lty))) {
         dirNext = nextDir0;
         iNext = contourMap->Index(ltx, lty);
      } else {
         AddNeighbor(ltx, lty, superSite);
         unsigned int stx(itx);
         unsigned int sty(contourMap->NextVisibleRow(itx, ity, delta1));
         if (superSiteSet.count(make_pair(stx, sty))) {
            dirNext = dirPrev;
            iNext = contourMap->Index(stx, sty);
         } else {
            AddNeighbor(stx, sty, superSite);
            dirNext = nextDir1;
            iNext = iPrev;
         }
      }
      break;
   }
   default:
      ASSERT(false);
   }
}

// Add a point to the boundary. Only add corners. The boundary is in
// actual display coordinates, taking into account hidden rows and
// columns, and a special inset to avoid drawing over the borders of
// neighboring tiles.
void TileChecker::AddCorner(size_t index,
                            Direction dirPrev,
                            Direction dirNext,
                            double inset,
                            CoordVec& boundary)
{
   double axi(contourMap->GetAdjustedX(index) +
              diagramEmitter.GetSmallZynqOffset());
   double ayi(contourMap->GetAdjustedY(index));
   // Collect corners of the perimeter and inset them like other logic.
   if (dirPrev != dirNext) {
      axi += inset;
      ayi += inset;

      if (dirNext == Down || dirPrev == Down) {
         axi += DiagramEmitter::boxWidth - 2.0 * inset;
      }
      if (dirNext == Left || dirPrev == Left) {
         ayi += DiagramEmitter::boxHeight - 2.0 * inset;
      }

      if (dirNext == Left && dirPrev == Up) {
         axi += DiagramEmitter::boxWidth;
      }
      if (dirNext == Up && dirPrev == Right) {
         ayi += DiagramEmitter::boxHeight;
      }
      if (dirNext == Right && dirPrev == Down) {
         axi -= DiagramEmitter::boxWidth;
      }
      if (dirNext == Down && dirPrev == Left) {
         ayi -= DiagramEmitter::boxHeight;
      }

      boundary.push_back(axi);
      boundary.push_back(ayi);
   }
}

// R#check_irs_separation, R#check_interconnect_routing - Is this tile
// outside all the isolated area ranges?
bool TileChecker::OutsideIsolatedRanges(size_t i) {
   size_t columns(contourMap->ColumnCount());
   size_t x(i % columns);
   size_t y(i / columns);
   bool isFence(true);
   for (Data::GroupIt gIt(data.groupData.begin());
        isFence && gIt != data.groupData.end();
        ++gIt)
   {
      if (gIt->first != "combined") {
         DeviceModel::SiteRectVec& rangeList(gIt->second->GetSiteRectList());
         for (DeviceModel::SiteRectVecIt rlIt(rangeList.begin());
              rlIt != rangeList.end();
              ++rlIt)
         {
            DeviceModel::SiteRect rc(*rlIt);
            if (deviceModel->IsUserTileType(rc.siteType)) {
               if (rc.Contains(x, y)) {
                  isFence = false;
                  goto out;
                  // Which is worse: multiple returns, an extra test
                  // and extra break, or a goto? ;-)
               }
            }
         }
      }
   }
  out:
   return isFence;
}

// Collect tiles outside all isolated area ranges for later
// reporting. Note: these are nominal fence tiles. Not all are
// necessarily valid fence tiles. Moreover, the area ranges are merely
// advisory. Isolated routing and logic may be located outside
// isolated area ranges, provided no actual isolation violation is
// formed.
void TileChecker::CollectNominalFenceTiles() {
   Classifier::NameSet nominals;
   for (size_t i(0); i < grid.size(); ++i) {
      if (OutsideIsolatedRanges(i)) {
         fenceList.push_back(i);
         Tile* tile(grid.at(i));
         if (tile) {
            tile->CollectNets(nominals);
         }
      }
   }
   nominalFenceNets = nominals.size();
}

// Collect coordinates of tiles inside super sites. These are
// typically null tiles in the device model. These coordinates are
// needed for finding the perimeters of super sites with irregular
// footprints such as Spartan-6 GTPs.
void TileChecker::CollectSuperSiteTileCoordinates() {
   for (size_t i(0); i < grid.size(); ++i) {
      Tile* tile(grid.at(i));
      if (tile) {
         Rf_Cell cell(tile->GetCell());
         if (!deviceModel->IsInterconnectCell(cell)) {
            // Stash the non-interconnect tiles so we can identify
            // supersite footprints for large tiles such as DSPs and
            // GTPs, etc.
            superSiteSet.insert(make_pair(cell.GetTileColumn(),
                                          cell.GetTileRow()));
         }
      }
   }
}

// Remove cells that do not belong to GTP sites from the
// superSiteSet. We need to do this because the only way to find the
// perimeter of the GTP is to subtract rectangular supersites which
// may be adjacent to the GTP.
//
// This is needed because the Spartan-6 PCIE site borders a GTP site,
// but there is no obvious line of demarcation in the device model.
// To resolve the ambiguity, architectural knowledge of the PCIE is
// used: the bounding box of its interconnect cells can be subtracted
// to yield the GTP footprint.
void TileChecker::RemoveCells(BoundInts& bb) {
   ASSERT(bb.initialized);
   // We don't actually need the +1 because the left and right columns
   // of the PCIE are INT cells that didn't get added to superSiteSet
   // in the first place, but we do it this way in case 7 series has
   // the INTs arranged differently.
   ASSERT(bb.bottom.GetTileRow() <= bb.top.GetTileRow());
   for (size_t x(bb.left.GetTileColumn());
        x != bb.right.GetTileColumn() + 1;
        ++x)
   {
      for (size_t y(bb.bottom.GetTileRow());
           y != bb.top.GetTileRow() + 1;
           ++y)
      {
         superSiteSet.erase(make_pair(x, y));
      }
   }
}

// Go one visible row or column outside the bounding box and note
// the neighbors for later adjacency testing.
//
// This is used for supersites that cannot be used as a fence.
void TileChecker::AddNeighborsForRectangle(BoundInts& bb,
                                           const Rf_CellDef& tileDef,
                                           const Rf_Cell superSite)
{
   ASSERT(bb.initialized);
   unsigned int left(bb.left.GetTileColumn());
   unsigned int top(bb.top.GetTileRow());
   unsigned int right(bb.right.GetTileColumn());
   unsigned int bottom(bb.bottom.GetTileRow());
   ASSERT(bottom <= top);
   ASSERT(left <= right);
   unsigned int box[4];

   box[Left]  = contourMap->NextVisibleColumn(left,  top,    -1);
   box[Up]    = contourMap->NextVisibleRow   (left,  top,     1);
   box[Right] = contourMap->NextVisibleColumn(right, bottom,  1);
   box[Down]  = contourMap->NextVisibleRow   (right, bottom, -1);

   // Since the coordinates are unsigned, underflow and overflow are
   // handled together. These checks are needed for the V5 site
   // PCIE_X0Y2 and for the Zynq site PS7_X0Y0.
   if (contourMap->ColumnCount() <= box[Left]) {
      box[Left] = left;
   }
   if (contourMap->RowCount() <= box[Down]) {
      box[Down] = bottom;
   }

   Rf_Cell startInt(deviceModel->GetIntTile(superSite, tileDef, 1));
   AddLogic al(contourMap, superSite);
   deviceModel->MapIntTiles(al, superSite, tileDef);

   // Add neighbors to every tile on the perimeter of the supersite.
   Tile* startTile(diagramEmitter.GetTile(startInt));

   GatherGroups gg(contourMap, groupSetFamily);
   deviceModel->MapIntTiles(gg, superSite, tileDef);

   // Hack to avoid start tile getting an X. The supersite gets
   // crosshatching so the start tile doesn't need an X.
   Data::GroupSet* gs(startTile->GetGroupSet());
   startTile->SetGroupSet(gg.GetGroupSet());

   const string siteType(superSite.GetCellDef().GetName());
   for (size_t x(box[Left]); x <= box[Right]; ++x) {
      AddNeighbor(x, box[Up], startTile);
      AddNeighbor(x, box[Down], startTile);
   }
   for (size_t y(bottom); y <= top; ++y) {
      AddNeighbor(box[Left], y, startTile);
      AddNeighbor(box[Right], y, startTile);
   }
   startTile->SetGroupSet(gs);
}

// The polygonal supersites are handled by
// CollectPolygonalBoundaries(). Here we mark the neighbors of the
// rectangular ones.
// This function was formerly FilterRectangularSuperSites.
void TileChecker::AddNeighborsForSuperSites() {
   Rf_CellDef tileDef(deviceModel->GetRfDevice()->GetTopCellDefTile());
   Rf_CellContainer sites(deviceModel->GetSites());

   // For each site, except the GTPs
   for (Rf_CellContainer::const_iterator sitesIt(sites.begin());
        sitesIt != sites.end();
        ++sitesIt)
   {
      // E.g. DSP48A1
      Rf_Cell site(*sitesIt);
      // Since the GTPs have non-rectangular footprints and since
      // there's no good way to identify cells with sites, we'll do it
      // by architectural knowledge. 
      const string siteType(site.GetCellDef().GetName());
      if (siteType == "PCIE_A1" || siteType == "PS7") {
         BoundInts bb;
         deviceModel->MapIntTiles(bb, site, tileDef);
         ASSERT(bb.initialized);
         RemoveCells(bb);
         AddNeighborsForRectangle(bb, tileDef, site);
      }
   }
}

// Find the perimeter of an irregularly shaped supersite such as the
// Spartan-6 GTPA1_DUAL. Updates superBoundaryMap.
void TileChecker::TraceContour(Rf_Cell site, Rf_CellDef tileDef) {
   // The INT cells of the supersite are on the perimeter.
   {
      // Add the INTs to the supersite set.
      AddInts ai(superSiteSet);
      deviceModel->MapIntTiles(ai, site, tileDef);

      AddCells ac(superIntSet);
      deviceModel->MapIntTiles(ac, site, tileDef);
   }
   Rf_Cell startInt(deviceModel->GetIntTile(site, tileDef, 1));
   long tx1(startInt.GetTileColumn()), ty1(startInt.GetTileRow());
   size_t iStart(contourMap->Index(tx1, ty1));
   ASSERT(superSiteSet.count(make_pair(tx1, ty1)));

   // We know this INT is in the supersite just inside the perimeter,
   // but we don't know a cell outside the perimeter that will tell us
   // which direction to follow the perimeter. Unfortunately, the GTP
   // site extends up above its top INT cell. However, its INTs appear
   // to be at the left and right edges in the examples we've looked at.

   long ty2(ty1);
   long tx2(contourMap->NextVisibleColumn(tx1, ty1, -1));
   Direction startDir(Down); // Assume right edge for now.
   if (!superSiteSet.count(make_pair(tx2, ty2))) {
      // The starting INT startInt is on the left edge of the site.
      startDir = Up;
   }

   size_t iPrev(iStart);
   Direction dirPrev(startDir);
   // Trivial upper bound on the length of a supersite perimeter.
   long duh(contourMap->RowCount() * contourMap->ColumnCount() * 4);
   // Store boundary information in superBoundaryMap and superGlobalMap.
   CoordVec& boundary(superBoundaryMap[site]);
   CoordVec& inset(superGlobalMap[site]);
   Tile* tile(diagramEmitter.GetTile(startInt));

   // Map over the INTs looking for globals.
   ScanForGlobals sfg(contourMap);
   deviceModel->MapIntTiles(sfg, site, tileDef);

   for (;duh;--duh) {
      Direction dirNew;
      size_t iNext;
      StepAlongPerimeter(tile, iPrev, dirPrev, iNext, dirNew);
      AddCorner(iNext, dirPrev, dirNew, DiagramEmitter::specialInset,
                boundary);
      if (sfg.hasGlobal) {
         AddCorner(iNext, dirPrev, dirNew, DiagramEmitter::globalInset, inset);
      }
      if (iNext == iStart && dirNew == startDir) {
         // Back at the start. Perimeter is complete.
         break;
      }
      dirPrev = dirNew; iPrev = iNext; 
   }
   // Failure here indicates that the loop exit condition isn't
   // right or isn't being met.
   ASSERT(duh);
}

// For all the non-rectangular supersites, find and store the
// boundaries.
void TileChecker::CollectPolygonalBoundaries() {
   Rf_CellDef tileDef(deviceModel->GetRfDevice()->GetTopCellDefTile());
   Rf_CellContainer sites(deviceModel->GetSites());

   for (Rf_CellContainer::const_iterator sitesIt(sites.begin());
        sitesIt != sites.end();
        ++sitesIt)
   {
      // E.g. DSP48A1
      Rf_Cell site(*sitesIt);
      // Infer tiles associated with a non-rectangular supersite as
      // follows:
      //    1. The interconnect tiles of a site are in the site.
      //    2. Sites are bounded by the edges of the device model.
      //    3. Rectangular sites are bounded by their interconnect
      //       tiles.
      //    4. Sites are contiguous.
      //    5. The interconnect tiles of a site are on the perimeter.
      // To find the perimeter of a supersite, start with a tile known
      // to be in the site and on the perimeter. Follow the perimeter
      // by identifying pairs of adjacent tile coordinates of which
      // one is not in the supersite (for any of the reasons above)
      // and the other is. Continue until the contour returns to the
      // starting point in the staring orientation.
      if (deviceModel->SpansMultipleTiles(site) && !IsRectangular(site)) {
         AddLogic al(contourMap, site);
         deviceModel->MapIntTiles(al, site, tileDef);
         // Walk the boundary of this site.
         TraceContour(site, tileDef);
      }
   }
}

// R#check_direct_routing, R#check_logic_tiles - Recognize blocks at
// the so-called 'comp' level. Comp is short for component. In this
// case, components are portions of a logic block or hard IP tile. For
// example, a CLB has comps for each of its slices. Comps are
// decomposed into 'bels' or basic elements. Bels correspond to
// netlist primitives such as AND3 or SRL16. Note that direct
// connections such as the carry chain are part of the configuration
// of the corresponding comp, so the absence of a comp implies the
// absence of direct connections.
//
// The existence of a comp for a specific physical location signals to
// BitGen to configure that location to behave as described by the
// comp and bels underneath it.
//
// The absence of a comp in a given location implies that that
// location is configured in its default state (virtually always a
// low-power inactive state).
struct IsCompFilter {
   bool operator()(const XCad_CadBlock* block) {
      return block->GetHierLevel() == XCad_CadBlock::LEVEL_COMP;
   }
};

// These clocking components do not contain user data (unless
//the user deliberately works to put it there), do not
//connect to the interconnect tiles in the usual way, and do
//span large regions of the chip. In short, including them is
//nothing but trouble.

// Note: this is related to the list in in
// Classifier::IsDrivenByRegionalClockBuffer(), but not exactly the
// same. We're using this pattern for sites (as opposed to nets) that
// do not present a risk of an isolation violation unless they are
// used in some crafty way beyond the capabilities of this automated
// analysis.
//
// IPAD is on the list because the code to identify the geometric
// extent of IPAD ranges does not work in all cases, however in all
// cases, any signals entering an IPAD will pass through another site
// that is checked, so there is no loss of coverage by excluding the
// IPAD site type.
const boost::regex
regionalClockPat("/?(?:BUFGCTRL|BUFGMUX|BUFH|BUFIO|BUFMR|BUFPLL|BUFR|"
                 "DCI|IDELAYCTRL|IPAD|SYSMON).*");

// R#check_irs_separation, R#check_direct_routing, R#check_logic_tiles -
// Collect all the comps so that all the tiles that are not in their
// default configuration can be identified. The complement of this set
// is the set of tiles in their unconfigured default states.
void TileChecker::MarkComps() {
   XCad_CadView* view(data.groupData.find("combined")->second->GetView());
   Rf_CellDef tileDef(deviceModel->GetRfDevice()->GetTopCellDefTile());
   Rf_CellContainer sites(deviceModel->GetSites());

   vector<XCad_CadBlock*> blocks;
   view->GetFilteredBlocks(blocks, IsCompFilter());

   // For all the comp blocks, find the placement and mark the
   // corresponding tile or tiles in the case of sites that occupy
   // more than one tile such as block RAM and DSPs.
   bool gotProgramming(false);
   size_t dotCounter(0);
   for (vector<XCad_CadBlock*>::iterator compIt(blocks.begin());
        compIt != blocks.end();
        ++compIt)
   {
      if (!dotCounter) ConsoleDot();
      if (dotCounter++ == 10000) dotCounter = 0;
      
      XCad_CadBlock* comp(*compIt);
      const Xdm_Property* prop(comp->GetPropertyByName("PDS_BLOCK_PLACEMENT"));
      if (prop) {
         const string siteNameByProp(prop->GetPropertyString());
         // These clocking components do not contain user data (unless
         //the user deliberately works to put it there), do not
         //connect to the interconnect tiles in the usual way, and do
         //span large regions of the chip. In short, including them is
         //nothing but trouble.
         if (boost::regex_match(siteNameByProp, regionalClockPat)) {
            continue;
         }
         for (Rf_CellContainer::const_iterator sitesIt(sites.begin());
              sitesIt != sites.end();
              ++sitesIt)
         {
            Rf_Cell site(*sitesIt);
            const string siteNameByCell(site.GetHierName());
            if (siteNameByProp == siteNameByCell) {
               // The Spartan-6 I/O architecture makes this analysis invalid.
               if (deviceModel->IsSpartan6() &&
                   string(site.GetCellDef().GetName()) == "PAD")
               {
                  continue;
               }

               // It is hard to imagine a real design without any
               // logic. Issue an internal error if none is found.
               gotProgramming = true;
               GatherGroups gg(contourMap, groupSetFamily);
               deviceModel->MapIntTiles(gg, site, tileDef);
               Data::GroupSet* gs(gg.GetGroupSet());
               Rf_Cell cell(deviceModel->GetSiteParentTile(site));
               Tile* tile(diagramEmitter.GetTile(cell));
               tile->AddCompName(siteNameByCell);
               tile->AddLogicSite(site);
               MarkOneComp moc(contourMap, site, gs, groupSetFamily);
               deviceModel->MapIntTiles(moc, site, tileDef);
               tile->SetGroupSet(gs);

               break;
            }
         }
      } else {
         const string& blockName(comp->GetUserHierName());
         Warn(120, "Unable to obtain placement information for block"
              + blockName + ".\n"
              "Verify that the design is fully placed and routed.");
      }
   }
   if (!gotProgramming) {
      FatalInternal(132, "Unable to find logic configuration information.");
   }
}

// Config Center logic tiles
const boost::regex configCenterTilePat(
   "(?:MONITOR_(?:TOP|MID|BOT).*)|(?:CFG_CENTER_(?:TOP|MID|BOT))|"
   "(?:CFG_SECURITY_(?:TOP|MID|BOT).*)");

// Config Center logic sites within the logic tiles
const boost::regex configCenterSitePat(
   "BSCAN|CAPTURE|DNA_PORT|EFUSE_USR|FRAME_ECC|ICAP|STARTUP|USR_ACCESS|XADC|"
   "CAPTURE|CFG_IO_ACCESS|DCIRESET|PMVIOB");

// For UCF-mode, use the isolated site ranges to populate
// the group sets of all the tiles.
//
// In addition to populating the sites of the user tiles, sites
// associated with the config center region area added so that
// violations can be reported for designs that use the config center
// area as a fence or designs that put config center sites in more
// than one isolation group.
void TileChecker::MapSitesToTiles() {
   Rf_CellDef tileDef(deviceModel->GetRfDevice()->GetTopCellDefTile());
   Rf_CellContainer sites(deviceModel->GetSites());

   // For each site
   for (Rf_CellContainer::const_iterator sitesIt(sites.begin());
        sitesIt != sites.end();
        ++sitesIt)
   {
      // E.g. DSP48A1
      Rf_Cell site(*sitesIt);
      string siteType(site.GetCellDef().GetName());

      // The CFG_IO_ACCESS site has no interconnect directly associated with
      // it, but Rf_TilePinHelpers makes one up causing a false
      // violation with a block RAM horizontally adjacent.
      //
      // This was confirmed by a detailed study of the
      // schematics and automated analysis using TraceToInt().
      if (!boost::regex_match(siteType, regionalClockPat) &&
          siteType != "CFG_IO_ACCESS")
      {
         // E.g. MACCSITE2
         Rf_Cell cell(deviceModel->GetSiteParentTile(site));
         string cellType(cell.GetCellDef().GetName());
         // Get the tile from the cell coordinates.
         Port::UINT32 column(cell.GetTileColumn());
         Port::UINT32 row(cell.GetTileRow());

         Tile* tile(contourMap->At(column, row));
         ASSERT(tile);

         // Since the XADC is directly connected to the
         // IPAD sites, including the IPADs only
         // produces spurious violations.
         if (!(cellType == "MONITOR_BOT_FUJI2" && siteType == "IPAD")) {

            // For each area range constraint
            for (Data::SiteRectMapIt srmIt(data.siteRectMap.begin());
                 srmIt != data.siteRectMap.end();
                 ++srmIt)
            {
               Data::IsolationGroupName group(srmIt->first);
               if (group == "<no isolation group>") continue;

               // For each range
               DeviceModel::SiteRectVec rangeList(srmIt->second);
               for (DeviceModel::SiteRectVecIt rlIt(rangeList.begin());
                    rlIt != rangeList.end();
                    ++rlIt)
               {
                  DeviceModel::SiteRect rc(*rlIt);
                  if (rc.UcfSiteEqual(siteType)) {

                     // The Spartan-6 I/O architecture makes this
                     // analysis invalid.
                     if (deviceModel->IsSpartan6() && siteType == "PAD") {
                        continue;
                     }

                     // SiteRect::Contains() only applies to individual
                     // interconnect tiles.  Several interconnect
                     // tiles may be used for a logic site.  To test a
                     // logic site for inclusion in an area range, it
                     // is necessary to check every interconnect tile
                     // associated with the logic tile for
                     // inclusion. For a tall logic site, a range may
                     // pass in between the top and the bottom.

                     bool isContained(false);
                     DeviceModel::CellSet ints;
                     AddCells ac(ints);
                     deviceModel->MapIntTiles(ac, site, tileDef);
                     for (DeviceModel::CellSetIt iIt(ints.begin());
                          iIt != ints.end();
                          ++iIt)
                     {
                        Rf_Cell intCell(*iIt);
                        Port::UINT32 x(intCell.GetTileColumn());
                        Port::UINT32 y(intCell.GetTileRow());
                        if (rc.Contains(x, y, siteType)) {
                           isContained = true;
                           break;
                        }
                     }
                     if (isContained) {

                        // Add the given group to the set of groups
                        // represented in this tile, if it is not already
                        // present.

                        Data::GroupSet* gs0(tile->GetGroupSet());
                        Data::GroupSet* gs1(
                           groupSetFamily.FindOrMakeSet(group, gs0));
                        tile->SetGroupSet(gs1);
                        tile->SetHasIsolatedNet();
                        tile->AddLogicSite(site);
                        MarkOneComp moc(contourMap, site, gs1,
                                        groupSetFamily);
                        moc.SetUcfMode();
                        deviceModel->MapIntTiles(moc, site, tileDef);
                        MapCmtCohort(group, tile, site);
                     }
                  }
               }
            }
         }
      }
   }
}

// The Spartan-6 clock sites are allowed in a horizontal fence, but
// not a vertical fence. To model this, for each neighbor to the
// immediate left or right of the DCM column, add neighbors on the
// opposite side of the DCM.
//
// Given a DCM site at (20, 118) extending up to (20, 119), all the
// labeled cells below are considered neighbors. No special treatment
// is needed for the cells directly above and below the DCM.
//
//      CLB        DCM       CLB
//   |---------+---------+---------|
//   | 19, 120 |         | 21, 120 |
//   |---------+---------+---------|
//   | 19, 119 | 20, 119 | 21, 119 |
//   |---------+---------+---------|
//   | 19, 118 | 20, 118 | 21, 118 |
//   |---------+---------+---------|
//   | 19, 117 |         | 21, 117 |
//   |---------+---------+---------|
//
// To continue the example, the code below will start at the bottom
// row, row 117 and proceed upwards. For each row, it will add
// one neighbor to the left and one neighbor to the right.
size_t TileChecker::CheckSpartan6ClockSite(Tile* tile) {
   ASSERT(tile);
   ASSERT(tile->HasSiteType("DCM") ||
          tile->HasSiteType("DCM_ADV") ||
          tile->HasSiteType("DCM_CLKGEN") ||
          tile->HasSiteType("DCM_SP") ||
          tile->HasSiteType("PLL_ADV") ||
          tile->HasSiteType("PLL_BASE"));
   Rf_Cell cell(tile->GetCell());
   size_t cellRow(cell.GetTileRow());
   size_t cellColumn(cell.GetTileColumn());
   size_t ri(contourMap->NextVisibleRow(cellColumn, cellRow, -1));

   for (int dy(0); dy < 4; ++dy) {        // dy =  0, 1, 2, 3
      for (int dx(-1); dx < 2; dx +=2) {  // dx = -1, 1
         unsigned int ci(contourMap->NextVisibleColumn(cellColumn, ri, dx));
         AddNeighbor(ci, ri, tile);
      }
      ri = contourMap->NextVisibleRow(cellColumn, ri, 1);
   }
   return 2;
}

// Spartan-6 has some special case fence rules. Of these only the DSP
// rule is implemented.
//    Two unused DSP sites must separate isolated DSPs from distinct
//    isolation groups.
// Note: the interconnect blocks associated with an unused DSP can
// be used without the risk that information passing through the 
// interconnect is communicated to the unused DSP.
size_t TileChecker::CheckS6SpecialSite(Tile* tile) {
   size_t rowsPerSite(0);
   ASSERT(tile);
   if (tile->HasSiteType("DCM") ||
       tile->HasSiteType("DCM_ADV") ||
       tile->HasSiteType("DCM_CLKGEN") ||
       tile->HasSiteType("DCM_SP") ||
       tile->HasSiteType("PLL_ADV") ||
       tile->HasSiteType("PLL_BASE"))
   {
      rowsPerSite = CheckSpartan6ClockSite(tile);
   } else if (tile->HasSiteType("DSP48A1")) {
      Rf_Cell logicSite(tile->GetLogicSite());
      // This is an interconnect tile.
      const size_t spartan6VerticalDspFence(2); // Units: sites.
      Rf_Cell cell(tile->GetCell());
      ASSERT(cell);
      const size_t column(cell.GetTileColumn());
      const size_t row(cell.GetTileRow());
      size_t rowAbove(row);
      // Skip rows that are part of the given logic site.
      Tile* top(contourMap->At(column, row)); // Needed for reporting.
      while (0 < rowAbove) {
         top = contourMap->At(column, rowAbove);
         if (!top ||
             top->logicSiteSet.size() != 1 ||
             top->GetLogicSite() != logicSite)
         {
            break;
         }
         --rowAbove;
         ++rowsPerSite;
      }
      // Test the two sites above, row-by-row.
      string violation;      // None yet.
      string sep;
      // Two unused DSPs are required, hence 2 * rowsPerSite.
      size_t scanCount(spartan6VerticalDspFence * rowsPerSite);
      for (size_t counter(0);
           0 < rowAbove && counter < scanCount;
           ++counter)
      {
         // Skip hidden rows.
         while (0 < rowAbove && contourMap->IsHiddenRow(rowAbove)) {
            --rowAbove;
         }
         Tile* above(contourMap->At(column, rowAbove));
         if (0 < rowAbove) {
            // If we decrement unconditionally and rowAbove was
            // zero beforehand, it will wrap to MAX_ULONG and
            // IsHiddenRow() attempt an invalid array access.
            --rowAbove;
         }
         // Conflicting comp?
         if (above->HasProgramming()
             && above->GetGroupSet()
             && tile->GetGroupSet() != above->GetGroupSet())
         {
            // There are conflicting groups.
            tile->SetHasBadNeighbor();

            // Mark all grid cells of the conflicting tile above.
            above->SetHasBadNeighbor();

            // Collect the violating tiles above.
            violation += sep
               + "    Group: " + SetToString(above->GetGroupSet())
               + " Site: " + string(above->GetLogicSite().GetName())
               + " Tile: " + above->GetName();
            sep = "\n";
         }
      }
      if (!violation.empty()) {

         // Mark all grid cells the site passed in.
         for (size_t i(0); i < rowsPerSite; ++i) {
            contourMap->At(column, row - i)->SetHasBadNeighbor();
         }

         string dspGroup(SetToString(tile->GetGroupSet())); 
         violation = "\n  Group: " + dspGroup
            + " DSP Site: "
            + string(logicSite.GetName()) + " Tiles: "
            + tile->GetName() + ":" + top->GetName()
            + "\n  is not isolated from the following sites:\n"
            + violation;
         s6DspFaultSet.insert(violation);
      }
   } else {
      rowsPerSite = 1;
   }
   return rowsPerSite;
}

// Check whether the Zynq PS7 site or any of the interconnect adjacent
// to it have neighbors of multiple isolation groups.
void TileChecker::CheckZynqPSS() {
   Rf_CellDef tileDef(deviceModel->GetRfDevice()->GetTopCellDefTile());
   Rf_CellContainer sites(deviceModel->GetSites());
   for (Rf_CellContainer::const_iterator sitesIt(sites.begin());
        sitesIt != sites.end();
        ++sitesIt)
   {
      Rf_Cell superSite(*sitesIt);
      const string siteType(superSite.GetCellDef().GetName());
      if (siteType == "PS7") {
         Rf_Cell cell(deviceModel->GetSiteParentTile(superSite));
         Tile* tile(diagramEmitter.GetTile(cell));

         Rf_Cell startInt(deviceModel->GetIntTile(superSite, tileDef, 1));
         AddLogic al(contourMap, superSite);
         deviceModel->MapIntTiles(al, superSite, tileDef);

         Tile* startTile(diagramEmitter.GetTile(startInt));

         if (1 < startTile->CountNeighborGroups()) {
            tile->SetHasBadNeighbor();
            TileSet neighbors(startTile->GetNeighbors());
            for (TileSetIt adjIt(neighbors.begin());
                 adjIt != neighbors.end();
                 ++adjIt)
            {
               Tile* neighbor(*adjIt);
               if (neighbor->HasGroups() && !tile->Equal(*neighbor)) {
                  neighbor->SetHasBadNeighbor();
                  if (neighbor->HasIsolatedNet()) {
                     netAdjacencyFaults.push_back(make_pair(tile, neighbor));
                  }
               }
            }
         }
      }
   }
}

// Config Center padding tiles
const boost::regex feedPat("(?:INT|HCLK)_FEEDTHRU_[12]|NULL");

// Ain't C++ great?
const Port::UINT32 uninitialized(static_cast<Port::UINT32>(-1));

// Find the boundaries of the Config Center area of a 7 series device.
//
// This is accomplished by looking for a particular pattern of tiles
// unique to the config center area.
void TileChecker::Find7SeriesConfigCenterBounds(Port::UINT32& columnMin,
                                                Port::UINT32& rowMin,
                                                Port::UINT32& columnMax,
                                                Port::UINT32& rowMax)
{
   columnMin = uninitialized;
   rowMin = uninitialized;
   columnMax = uninitialized;
   rowMax = uninitialized;

   // The actual config center tiles occupy an area spanning the
   // middle of the right edge of the config center region. Their
   // coordinates do not give us the extrema, but do help skip a null
   // row across the middle.

   // Use config center tiles to get preliminary bounds.
   Rf_CellContainer cells(deviceModel->GetCells());
   for (Rf_CellContainer::const_iterator i(cells.begin());
        i != cells.end();
        ++i)
   {
      Rf_Cell cell(*i);
      const string cellTypeName(cell.GetCellDef().GetName());
      if (boost::regex_match(cellTypeName, configCenterTilePat)) {
         Port::UINT32 row(cell.GetTileRow());

         // Update the row bounds.
         if (rowMin == uninitialized || row < rowMin) {
            rowMin = row;
         }
         if (rowMax == uninitialized || rowMax < row) {
            rowMax = row;
         }

         // Update the right edge. The matching cell is on the right
         // edge.
         if (columnMax == uninitialized) {
            columnMax = cell.GetTileColumn() - 1;
         } else {
            ASSERT(columnMax == cell.GetTileColumn() - 1);
         }

         // Scan to the left for the left edge. The Config Center area
         // is delimited on the left by a VBRK tile. 
         for (Port::UINT32 column(columnMax);
              0 < column;
              --column)
         {
            Tile* tile(contourMap->At(column, row));
            if (tile) {
               const string typeName(tile->GetCell().GetCellDef().GetName());
               if (typeName == "VBRK") {
                  if (columnMin == uninitialized) {
                     columnMin = column + 1;
                  } else {
                     ASSERT(columnMin == column + 1);
                  }
                  break;
               }
            }
         }
      }
   }
   ASSERT(columnMin != uninitialized);
   ASSERT(columnMax != uninitialized);
   ASSERT(rowMin != uninitialized);
   ASSERT(rowMax != uninitialized);
   // Use knowledge of the tile types to find the top edge of the
   // config center region. Note: this is the interior of the region;
   // a further increment will be needed to find the top neighbors.
   Port::UINT32 scanColumn(columnMax - 1);
   for (Port::UINT32 row(rowMin); 0 < rowMin; --row) {
      Tile* tile(contourMap->At(scanColumn, row));
      if (tile) {
         const string typeName(tile->GetCell().GetCellDef().GetName());
         if (!boost::regex_match(typeName, feedPat)) {
            rowMin = row + 1;
            break;
         } else if (!row) {
            rowMin = 0;
            break;
         }
      }
   }
   // Find the bottom edge of the config center region.
   for (Port::UINT32 row(rowMax); row < contourMap->RowCount(); ++row) {
      Tile* tile(contourMap->At(scanColumn, row));
      if (tile) {
         const string typeName(tile->GetCell().GetCellDef().GetName());
         rowMax = row - 1;
         if (!boost::regex_match(typeName, feedPat)) {
            break;
         }
      }
   }
}

// Scan left from columnStart to get the neighboring interconnect
// column. This will be associated with a neighbor of the config
// center area, not config center itself.
Port::UINT32 TileChecker::FindLeftInterconnect(Port::UINT32 columnStart,
                                               Port::UINT32 row)
{
   Port::UINT32 result(uninitialized);
   for (Port::UINT32 column(columnStart); 0 < column; --column) {
      Tile* tile(contourMap->At(column, row));
      if (tile) {
         if (deviceModel->IsInterconnectCell(tile->GetCell())) {
            result = column;
            break;
         }
      }
   }
   return result;
}

// Scan right from columnMax to get the adjacent interconnect
// column. The first one we find is associated with the config center
// region itself. The second one is the neighbor.
Port::UINT32 TileChecker::FindRightInterconnect(Port::UINT32 columnStart,
                                                Port::UINT32 row)
{
   Port::UINT32 result(uninitialized);
   for (Port::UINT32 column(columnStart + 1);
        column < contourMap->ColumnCount();
        ++column)
   {
      Tile* tile(contourMap->At(column, row));
      if (tile) {
         if (deviceModel->IsInterconnectCell(tile->GetCell())) {
            result = column;
            break;
         }
      }
   }
   return result;
}

// Gather the groups occupied by tiles in the Config Center area by
// scanning the portion of the column containing the interconnect blocks
// of the config center tiles.
//
// The resulting group set is called configCenterGroupSet.
// The interconnect tiles are inserted into configCenterSet.
void TileChecker::Gather7SeriesConfigCenterGroups(
   Port::UINT32 configIntColumn,
   Port::UINT32 rowMin,
   Port::UINT32 rowMax)
{
   ASSERT(configIntColumn < contourMap->ColumnCount());
   ASSERT(rowMin < rowMax);
   for (Port::UINT32 row(rowMin); row < rowMax; ++row) {
      Tile* ccTile(contourMap->At(configIntColumn, row));
      if (ccTile) {
         if (ccTile->HasGroups()) {
            configCenterGroupSet =
               groupSetFamily.FindOrMakeUnion(configCenterGroupSet,
                                              ccTile->GetGroupSet());
            configCenterSet.insert(ccTile);
         }
      }
   }
}

// Record the interconnect tiles adjacent to the Config Center area in
// configCenterNeighborSet.
// Record the groups occupied by the neighbors in
// configCenterNeighborGroupSet.
void TileChecker::Add7SeriesConfigCenterNeighbor(Port::UINT32 column,
                                                 Port::UINT32 row)
{
   Tile* tile(contourMap->At(column, row));
   if (tile) {
      if (deviceModel->IsInterconnectCell(tile->GetCell())) {
         if (tile->HasGroups() || tile->HasTaints()) {
            configCenterNeighborSet.insert(tile);
            configCenterNeighborGroupSet =
               groupSetFamily.FindOrMakeUnion(configCenterNeighborGroupSet,
                                              tile->GetGroupSet());
            configCenterNeighborGroupSet =
               groupSetFamily.FindOrMakeUnion(configCenterNeighborGroupSet,
                                              tile->GetTaintGroupSet());
         }
      }
   }
}

// The config center area in 7 series cannot be used as a
// fence. Unfortunately, the device model does not represent this area
// in a way that it can be represented as the interconnect blocks of
// one site or a collection of sites. Instead, it is necessary to map
// out the region based on its user tiles and boundary tile types.
void TileChecker::Check7SeriesConfigCenter() {
   Port::UINT32 columnMin;
   Port::UINT32 rowMin;
   Port::UINT32 columnMax;
   Port::UINT32 rowMax;
   Find7SeriesConfigCenterBounds(columnMin, rowMin, columnMax, rowMax);

   // The interconnects of config center are on the right of tiles
   // used to identify its bounds. Expand to the right to pick up the
   // config center interconnects.
   columnMax = FindRightInterconnect(columnMax, rowMax);

   Gather7SeriesConfigCenterGroups(columnMax, rowMin, rowMax);

   // At this point columnMin, columnMax, rowMin, and rowMax bound the
   // config center region, but we haven't expanded out to interconnects.
   columnMin = FindLeftInterconnect(columnMin, rowMax);

   // We'll need the row min and max later for scanning the config
   // center column outside the config center area.
   configCenterRowMin = rowMin;
   configCenterRowMax = rowMax;

   // Expand to the right again to pick up the config center neighbor
   // interconnects.
   columnMax = FindRightInterconnect(columnMax, rowMax);

   size_t clockIntColumn;
   bool adjacent(false);
   {                            // From SweepSuperSites()
   
      Rf_CellDef tileDef(deviceModel->GetRfDevice()->GetTopCellDefTile());
      Rf_CellContainer sites(deviceModel->GetSites());

      Rf_Cell clockSite;
      Rf_Cell configCenterSite;

      // For each site
      for (Rf_CellContainer::const_iterator sitesIt(sites.begin());
           sitesIt != sites.end();
           ++sitesIt)
      {
         Rf_Cell site(*sitesIt);
         string siteType(site.GetCellDef().GetName());
         // Find the config center column and the clock buffer column.
         if (deviceModel->Is7Series()) {
            // Collect a clock site and config center site for a special
            // column sweep below.
            if (!clockSite &&
                boost::regex_match(siteType, deviceModel->clockBufferPat))
            {
               clockSite = site;
            }
            if (!configCenterSite &&
                boost::regex_match(siteType, deviceModel->configCenterPat))
            {
               configCenterSite = site;
            }
         }
      }
      if (deviceModel->Is7Series()) {
         Rf_Cell ccCell(deviceModel->GetSiteParentTile(configCenterSite));
         Port::UINT32 ccColumn(ccCell.GetTileColumn());
         Port::UINT32 ccRow(ccCell.GetTileRow());

         size_t ccIntColumn(
            contourMap->NextVisibleColumn(ccColumn, ccRow, 1));

         Rf_Cell clockCell(deviceModel->GetSiteParentTile(clockSite));
         Port::UINT32 clockColumn(clockCell.GetTileColumn());
         Port::UINT32 clockRow(clockCell.GetTileRow());

         clockIntColumn = (
            contourMap->NextVisibleColumn(clockColumn, clockRow, 1));

         // The delta between ccIntColumn and clockIntColumn can be seen
         // by examining the tile maps.
         adjacent = (ccIntColumn - clockIntColumn < 7 ||
                     clockIntColumn - ccIntColumn < 7);

      }
   }
   if (adjacent) {
      columnMax = clockIntColumn;
   }

   // We avoid checking the corners twice, though it doesn't make a
   // difference because the set would do that for us.

   // Scan left and right edge neighbors.
   for (Port::UINT32 row(rowMin); row <= rowMax; ++row) {
      Add7SeriesConfigCenterNeighbor(columnMin, row);
      Add7SeriesConfigCenterNeighbor(columnMax, row);
   }
   // Scan top and bottom edge neighbors, unless the top or bottom
   // coincides with the top or bottom edges of the device.
   for (Port::UINT32 column(columnMin); column <= columnMax; ++column) {
      if (2 < rowMin) {
         Add7SeriesConfigCenterNeighbor(column, rowMin - 2);
      }
      if (diagramEmitter.GetSmallZynqOffset() < 1.0) {
         Add7SeriesConfigCenterNeighbor(column, rowMax + 2);
      }
   }

   // Now we need to know, does this region use more than one
   // isolation group internally or when combined with its neighbors?
   bool hasNonMatchingNeighbors(
      configCenterGroupSet && configCenterNeighborGroupSet &&
      configCenterGroupSet != configCenterNeighborGroupSet);
   bool hasMultipleNeighbors(
      configCenterNeighborGroupSet &&
      1 < configCenterNeighborGroupSet->size());
   if (hasNonMatchingNeighbors || hasMultipleNeighbors) {
      for (TileSetIt adjIt(configCenterNeighborSet.begin());
           adjIt != configCenterNeighborSet.end();
           ++adjIt)
      {
         Tile* tile(*adjIt);
         if (tile->HasGroups() &&
             tile->GetGroupSet() != configCenterGroupSet)
         {
            tile->SetHasBadNeighbor();
            configCenterFaultSet.insert(tile);
         }
      }
   }
   bool hasMultipleGroups(
      configCenterGroupSet && 1 < configCenterGroupSet->size());
   if (hasMultipleGroups) {
      for (TileSetIt ccsIt(configCenterSet.begin());
           ccsIt != configCenterSet.end();
           ++ccsIt)
      {
         Tile* tile(*ccsIt);
         if (tile->HasGroups()) {
            tile->SetHasBadNeighbor();
            configCenterFaultSet.insert(tile);
         }
      }
   }
   configCenterFaultCount = configCenterFaultSet.size();
   if (configCenterFaultCount) {
      diagramEmitter.SetHasConfigCenterViolation();
   }
}

// Set a flag in data to indicate that a Zynq PS7 site is in the
// device but was not included in an isolated area range as required
// by the isolation methodology.
void TileChecker::CheckPS7Constraint() {
   // See if a PS7 constraint is present.
   data.omittedPS7 = deviceModel->IsZynq();
   if (data.omittedPS7) {
      // For each site range...
      for (Data::SiteRectMapIt srmIt(data.siteRectMap.begin());
           srmIt != data.siteRectMap.end();
           ++srmIt)
      {
         DeviceModel::SiteRectVec rangeList(srmIt->second);
         for (DeviceModel::SiteRectVecIt rlIt(rangeList.begin());
              rlIt != rangeList.end();
              ++rlIt)
         {
            DeviceModel::SiteRect rc(*rlIt);
            if (rc.siteType == "PS7") {
               data.omittedPS7 = false;
               break;
            }
         }
      }
   }
}

// The 7-series GTX transceivers have a special isolation rule: at
// most one isolation group per bank. The bank information is not
// easily accessible through the device model, however the regions
// correspond to groups of four GTXs. The y-coordinate mod 4
// partitions them into their I/O bank cohorts.
void TileChecker::Check7SeriesGTXE2() {
   for (size_t x(0); x < contourMap->ColumnCount(); ++x) {
      for (size_t y(0); y < contourMap->RowCount(); ++y) {
         Tile* tile(contourMap->At(x, y));
         if (tile) {
            string tileType(tile->GetCell().GetCellDef().GetName());
            Tile* tRef(tile->HasGroups() ? tile : 0);
            if (tileType == "GTX_CHANNEL_3") {
               for (size_t i(0); i < 3; ++i) {
                  string cohort("GTX_CHANNEL_" + ToString(i));
                  size_t yi(y + 1);
                  while (!contourMap->At(x, yi) ||
                         (string(contourMap->At(x, yi)->GetCell().GetCellDef().GetName()) != cohort && 
                          string(contourMap->At(x, yi)->GetCell().GetCellDef().GetName()) != "GTX_COMMON"))
                  {
                     ++yi; // Scan downward.
                  }
                  Tile* tNew(contourMap->At(x, yi));
                  if (tNew->HasGroups()) {
                     if (tRef) {
                        // There are two ways for multiple groups to
                        // be represented in a GTX cohort. Multiple
                        // cohorts can occupy distinct groups or a
                        // single tile can contain occupy multiple
                        // isolation groups. The latter case is
                        // covered elsewhere, so to avoid confusion,
                        // it is not reiterated here.
                        if (tRef->GetGroupSet() != tNew->GetGroupSet()) {
                           gtxFaultSet.insert(
                                  "Group: {" + SetToString(tRef->GetGroupSet())
                                  + "} tile: " + tRef->GetName() +
                                  "\n    is not isolated from group: {" +
                                  SetToString(tNew->GetGroupSet()) + "} tile: "
                                  + tNew->GetName());
                        } else if (1 < tRef->GetGroupSet()->size()) {
                           ; // A tile in multiple isolation groups
                             // will reported elsewhere...
                        }
                     } else {
                        tRef = tNew;
                     }
                  }
               }
            }
         }
      }
   }
   gtxFaultCount = gtxFaultSet.size();
}

// R#define_fence -
// Special fence rules have the following form:
//
//    IF a tile of particular type is used by an isolation group,
//  THEN some other particular tiles cannot be used by any other
//       isolation group.
//
void TileChecker::CheckSpecialSites() {
   if (deviceModel->IsSpartan6()) {
      for (size_t x(0); x < contourMap->ColumnCount(); ++x) {
         size_t dy(1);
         for (size_t y(contourMap->RowCount() - 1);
              y < contourMap->RowCount(); // Oh, joy: unsigned arithmetic!
              y -= dy)
         {
            Tile* tile(contourMap->At(x, y));
            if (tile) {
               // Since sites may occupy multiple tile rows, allow
               // the function below to determine how many rows
               // were processed.
               dy = CheckS6SpecialSite(tile);
            } else {
               dy = 1;
            }
         }
      }
      dspFaultCount = s6DspFaultSet.size();
   } else if (deviceModel->Is7Series()) {
      CheckPS7Constraint();
      Check7SeriesConfigCenter();
      if (deviceModel->IsZynq()) {
         CheckZynqPSS();
      }
      Check7SeriesGTXE2();
   }
}

// Add a range violation message to rangeViolationSet. We store the 
// message rather than the parameters so that the std::set will do the
// deduping and sorting. 
void TileChecker::AddRangeViolation(const Data::IsolationGroupName& g1,
                                    const DeviceModel::SiteRect& rc1,
                                    const Data::IsolationGroupName& g2,
                                    const DeviceModel::SiteRect& rc2,
                                    const string& violation)
{
   string v(
      "Isolation Violation:63: Isolation groups "
      + g1 + " and " + g2 + " " + violation + ".\n"
      "  Range " + rc1.rangeString + " from area group " + rc1.areaGroupName
      + " conflicts with\n"
      "  range " + rc2.rangeString + " from area group " + rc2.areaGroupName
      + ".");
   rangeViolationSet.insert(v);
}

// The vector netAdjacencyFaults has isolation violations represented
// by pairs <tile1, tile2> where tile1 != tile 2. We want to report
// pairs:
//    <<siteRange1, areaGroup1, isolationGroup1>, 
//     <siteRange2, areaGroup2, isolationGroup2>>, 
// where 
//    tile1 in siteRange1, 
//    tile2 in siteRange2,
//    siteRange1.site == siteRange2.site and
//    isolationGroup1 != isolationGroup2.
void TileChecker::CheckAdjacentAreaRanges(Tile* t1, Tile* t2) {
   Port::UINT32 x1(t1->GetCell().GetTileColumn());
   Port::UINT32 y1(t1->GetCell().GetTileRow());
   Port::UINT32 x2(t2->GetCell().GetTileColumn());
   Port::UINT32 y2(t2->GetCell().GetTileRow());

   // For each isolation group g1 occupied by t1
   Data::GroupSet* gs1(t1->GetGroupSet());
   Data::GroupSet* gs2(t2->GetGroupSet());
   if (gs1 && gs2) {
      for (Data::GroupSetIt g1It(gs1->begin());
           g1It != gs1->end();
           ++g1It)
      {
         Data::IsolationGroupName g1(*g1It);
         // For each isolation group g2 occupied by t2
         Data::GroupSet* gs2(t2->GetGroupSet());
         for (Data::GroupSetIt g2It(gs2->begin());
              g2It != gs2->end();
              ++g2It)
         {
            Data::IsolationGroupName g2(*g2It);
            if (g1 != g2) {
               // The sets gs1 and gs2 may contain multiple groups. When
               // g1 == g2, there is obviously no violation.

               // For each siteRect containing t1
               Data::SiteRectMapIt srm1It(data.siteRectMap.find(g1));
               if (srm1It != data.siteRectMap.end()) {
                  DeviceModel::SiteRectVec rangeList(srm1It->second);
                  for (DeviceModel::SiteRectVecIt rl1It(rangeList.begin());
                       rl1It != rangeList.end();
                       ++rl1It)
                  {
                     DeviceModel::SiteRect rc1(*rl1It);
                     for (DeviceModel::CellSetIt
                             ls1It(t1->logicSiteSet.begin());
                          ls1It != t1->logicSiteSet.end();
                          ++ls1It)
                     {
                        Rf_Cell site1(*ls1It);
                        const string siteType1(site1.GetCellDef().GetName());
                        if (!boost::regex_match(siteType1, regionalClockPat) &&
                            rc1.Contains(x1, y1, siteType1))
                        {
                           // For each siteRect containing t2
                           Data::SiteRectMapIt
                              srm2It(data.siteRectMap.find(g2));
                           if (srm2It != data.siteRectMap.end()) {
                              DeviceModel::SiteRectVec rangeList(srm2It->second);
                              for (DeviceModel::SiteRectVecIt
                                      rl2It(rangeList.begin());
                                   rl2It != rangeList.end();
                                   ++rl2It)
                              {
                                 DeviceModel::SiteRect rc2(*rl2It);
                                 for (DeviceModel::CellSetIt
                                         ls2It(t2->logicSiteSet.begin());
                                      ls2It != t2->logicSiteSet.end();
                                      ++ls2It)
                                 {
                                    Rf_Cell site2(*ls2It);
                                    const string
                                       siteType2(site2.GetCellDef().GetName());

                                    if (!boost::regex_match(siteType2,
                                                            regionalClockPat)
                                        && rc2.Contains(x2, y2, siteType2))
                                    {
                                       // Special case for Zynq PSS:
                                       // The PLL adjacent to the
                                       // bottom left of the processor
                                       // can be used without an
                                       // isolation violation. Of
                                       // course, to use it, it is
                                       // also necessary to use its
                                       // I/O FIFOs.
                                       if (!(siteType1 == "PS7" &&
                                             (siteType2 == "IN_FIFO" ||
                                              siteType2 == "OUT_FIFO" ||
                                              siteType2 == "PLLE2_ADV")))
                                       {
                                          // Record violation
                                          const string
                                             ref("are adjacent near " +
                                                 (siteType1 == "PS7"
                                                  ? string("PS7_X0Y0")
                                                  : t1->GetName()));
                                          AddRangeViolation(g1, rc1,
                                                            g2, rc2,
                                                            ref);
                                       }
                                    }
                                 }
                              }
                           }
                        }
                     }
                  }
               }
            }
         }
      }
   }
}

// We also want to report this same information for individual tiles
// in multiple groups as indicated by tile->HasBadContent().
void TileChecker::CheckOverlappingAreaRanges(Tile* tile) {
   ASSERT(tile);
   Port::UINT32 x(tile->GetCell().GetTileColumn());
   Port::UINT32 y(tile->GetCell().GetTileRow());

   // For each group the tile is in
   Data::GroupSet* gs1(tile->GetGroupSet());
   if (gs1) {
      for (Data::GroupSetIt gs1It(gs1->begin());
           gs1It != gs1->end();
           ++gs1It)
      {
         Data::IsolationGroupName g1(*gs1It);
         // For each siteRect containing tile
         Data::SiteRectMapIt srm1It(data.siteRectMap.find(g1));
         if (srm1It != data.siteRectMap.end()) {
            DeviceModel::SiteRectVec rangeList(srm1It->second);
            for (DeviceModel::SiteRectVecIt rl1It(rangeList.begin());
                 rl1It != rangeList.end();
                 ++rl1It)
            {
               DeviceModel::SiteRect rc1(*rl1It);

               for (DeviceModel::CellSetIt
                       lsIt(tile->logicSiteSet.begin());
                    lsIt != tile->logicSiteSet.end();
                    ++lsIt)
               {
                  Rf_Cell site(*lsIt);
                  const string siteType(site.GetCellDef().GetName());
                  if (rc1.Contains(x, y, siteType)) {
                     // For each group the tile is in after the current group
                     Data::GroupSetIt gs2It(gs1It);
                     for (++gs2It; gs2It != gs1->end(); ++gs2It) {
                        Data::IsolationGroupName g2(*gs2It);

                        // For each siteRect containing t2
                        Data::SiteRectMapIt srm2It(data.siteRectMap.find(g2));
                        if (srm2It != data.siteRectMap.end()) {
                           DeviceModel::SiteRectVec rangeList(srm2It->second);
                           for (DeviceModel::SiteRectVecIt
                                   rl2It(rangeList.begin());
                                rl2It != rangeList.end();
                                ++rl2It)
                           {
                              DeviceModel::SiteRect rc2(*rl2It);
                              if (rc2.Contains(x, y)) {
                                 // Record violation
                                 AddRangeViolation(g1, rc1, g2, rc2,
                                                   "overlap");
                              }
                           }
                        }
                     }
                  }
               }
            }
         }
      }
   }
}

// The FPGA designer is likely to run into difficulties if certain CMT
// sites are isolated without certain other sites also being isolated.

// 7 series I/O FIFOs and other site types in the CMT columns.
const boost::regex ioFifoPat(
   "IN_FIFO|OUT_FIFO|MMCME2_ADV|PLLE2_ADV|PHASER_IN_PHY|PHASER_OUT_PHY|"
   "PHY_CONTROL|PHASER_REF");

// Same as above without the I/O fifos.
const boost::regex cmtLogicPat(
   "MMCME2_ADV|PLLE2_ADV|PHASER_IN_PHY|PHASER_OUT_PHY|PHY_CONTROL|"
   "PHASER_REF");

// 7 series phaser in and out phys.
const boost::regex phaserPhyPat("PHASER_IN_PHY|PHASER_OUT_PHY");

// Mapping from isolation group to sets of site names.
CohortGroupMap isolatedCmtCohort;

const boost::regex phyPat("PHASER_(IN|OUT)_PHY_X(.*)Y(.*)");
const boost::regex prpcPat("(PHASER_REF|PHY_CONTROL)_X(.*)Y(.*)");

// Take note of the 7-series phaser-related site isolated in the given
// group.
void TileChecker::MapCmtCohort(const string group, Tile* tile, Rf_Cell site) {
   const string siteName(site.GetName());
   if (boost::regex_match(siteName, phyPat) ||
       boost::regex_match(siteName, prpcPat))
   {
      Tile* keyTile(GetKeyTile(tile));
      ASSERT(isolatedCmtCohort[group].find(siteName) ==
             isolatedCmtCohort[group].end());
      isolatedCmtCohort[group][siteName] = keyTile;
   }
}

// This tile is used as the canonical representative of a pair of
// related tiles, CMT_TOP_[LR]_UPPER_B and CMT_TOP_[LR]_LOWER_T that
// contain 6 and 4 related phaser sites, respectively.
const boost::regex keyTilePat("CMT_TOP_[LR]_UPPER_B");

// Get a canonical representation of a 7-series CMT phaser cohort.
Tile* TileChecker::GetKeyTile(Tile* tile) {
   Tile* result(tile);
   Rf_Cell cell(tile->GetCell());
   Port::UINT32 column(cell.GetTileColumn());
   Port::UINT32 row(cell.GetTileRow());
   while (!result ||
          !boost::regex_match(string(cell.GetCellDef().GetName()), keyTilePat))
   {
      ASSERT((row - 1) < row);  // Unsigned underflow.
      --row;
      result = contourMap->At(column, row);
      if (result) {
         cell = result->GetCell();
      }
   }
   return result;
}

// Helper function for CheckImpliedAreaRanges() below. This function
// is called when any phaser site occurs in an isolated area range. It
// inserts 10 entries into the relatedSites map, all of which must be
// in the isolation group of the site found at Y=n.
//
// These sites span two tiles.
//    Tile CMT_TOP_[LR]_UPPER_B contains
//        Site PHASER_IN_PHY_XxY(4*y + 3)
//        Site PHASER_OUT_PHY_XxY(4*y + 3)
//        Site PHASER_IN_PHY_XxY(4*y + 2)
//        Site PHASER_OUT_PHY_XxY(4*y + 2)
//        Site PHASER_REF_XxYy
//        Site PHY_CONTROL_XxYy
//    Tile CMT_TOP_[LR]_LOWER_T
//        Site PHASER_IN_PHY_XxY(4*y)
//        Site PHASER_OUT_PHY_XxY(4*y)
//        Site PHASER_IN_PHY_XxY(4*y + 1)
//        Site PHASER_OUT_PHY_XxY(4*y + 1)
void TileChecker::InsertImpliedPhaserSites(SiteNameTileMap& relatedSites,
                                           const string& xStr,
                                           size_t n,
                                           Tile* tile)
{
   // Use the PLL as the canonical tile for this group of cohorts.
   // The PLL is at the top of the three tiles.
   ASSERT(tile);
   Tile* keyTile(GetKeyTile(tile));

   const string ref("PHASER_REF_X" + xStr + "Y" + ToString(n));
   const string control("PHY_CONTROL_X" + xStr + "Y" + ToString(n));
   relatedSites.insert(make_pair(ref, keyTile));
   relatedSites.insert(make_pair(control, keyTile));
   for (size_t i(0); i < 4; ++i) {
      size_t yMod(4 * n + i);
      const string in("PHASER_IN_PHY_X" + xStr + "Y" + ToString(yMod));
      const string out("PHASER_OUT_PHY_X" + xStr + "Y" + ToString(yMod));
      relatedSites.insert(make_pair(in, keyTile));
      relatedSites.insert(make_pair(out, keyTile));
   }
}

// Compose a message of the form:
//    Group BECKETT
//        Sites: ESTRAGON, VLADIMIR
//        requires site: GODOT
// where the upper case words are specific logic sites.
const string TileChecker::ComposeImpliedViolation(
   TileSet sourceTiles,
   const string group,
   DeviceModel::CellSet impliedSet) const
{
   DeviceModel::CellSet sourceSites;
   for (TileSetIt tsIt(sourceTiles.begin());
        tsIt != sourceTiles.end();
        ++tsIt)
   {
      Tile* tile(*tsIt);
      sourceSites.insert(tile->logicSiteSet.begin(),
                         tile->logicSiteSet.end());
   }

   for (DeviceModel::CellSetIt csIt(impliedSet.begin());
        csIt != impliedSet.end();
        ++csIt)
   {
      sourceSites.erase(*csIt);
   }

   string source( 1 < sourceSites.size() ? "\n    Sites: " : "\n    Site: ");
   string sourceNames(SetToString(sourceSites));
   string require(1 < sourceSites.size() ? "\n    require" : "\n    requires");
   string missing(1 < impliedSet.size()  ? " sites: "      : " site: ");
   string missingNames(SetToString(impliedSet));
   return "\n" + WordWrap(group + source + sourceNames +
                          require + missing + missingNames);
}

// We believe that failure to include IN_FIFO and OUT_FIFO sites in
// isolated area ranges will lead to implementation failures. It's
// better to insist that the customer include these for each CMT site
// that uses them, specifically
//    MMCME2_ADV, PLLE2_ADV, PHASER_IN_PHY, PHASER_OUT_PHY,
//    PHY_CONTROL, and PHASER_REF
//
// The association between I/O FIFOs and other CMT sites is tedious
// to calculate. The I/O FIFOs associated with a CMT site are the ones 
// that use interconnect in a range that overlaps those used by the
// CMT.
//
// To find these, a map is constructed that supports the interconnect
// range query. Actually, two maps. The CmtMap stores a map for each
// column containing CMT tiles. The CmtColumnMap stores information
// for one column indexed by the row of the lowest interconnect needed
// by a site. For a given row, there may be several sites, so a row
// index maps to a set of records which include the interconnect range
// along with the CMT site and tile.
//
// Phaser components must be used in pairs.
//
// Create the CMT map. This tells us which CMT sites are related based
// on the interconnect tiles they use.
//
// Also, note any CMT tiles that are used but not isolated,
// i.e. contain top-level logic.
void TileChecker::MakeCmtMap(CmtMap& cmtMap) {
   ASSERT(deviceModel->Is7Series());
   Rf_CellDef tileDef(deviceModel->GetRfDevice()->GetTopCellDefTile());
   Rf_CellContainer sites(deviceModel->GetSites());
   for (Rf_CellContainer::const_iterator sitesIt(sites.begin());
        sitesIt != sites.end();
        ++sitesIt)
   {
      // E.g. DSP48A1
      Rf_Cell site(*sitesIt);
      string siteType(site.GetCellDef().GetName());
      if (deviceModel->IsUserTileType(siteType)) {
         // E.g. MACCSITE2
         Rf_Cell cell(deviceModel->GetSiteParentTile(site));
         string cellType(cell.GetCellDef().GetName());
         // Get the tile from the cell coordinates.
         Port::UINT32 column(cell.GetTileColumn());
         Port::UINT32 row(cell.GetTileRow());
         Tile* tile(contourMap->At(column, row));
         ASSERT(tile);

         if (boost::regex_match(siteType, ioFifoPat)) {
            BoundInts bb;
            deviceModel->MapIntTiles(bb, site, tileDef);
            ASSERT(bb.initialized);
            bb.tile = tile;
            bb.site = site;
            cmtMap[column][bb.bottom.GetTileRow()].insert(bb);
         }
      }
   }
}

// Identify CMT sites for which I/O FIFOs are not members of the CMTs
// isolation group.
void TileChecker::CheckImpliedIOFifos(CmtMap& cmtMap) {
   ASSERT(deviceModel->Is7Series());
   for (CmtMap::iterator mapIt(cmtMap.begin());
        mapIt != cmtMap.end();
        ++mapIt)
   {
      size_t col(mapIt->first);
      CmtColumnMap& colMap(mapIt->second);
      for (CmtColumnMap::iterator colIt(colMap.begin());
           colIt != colMap.end();
           ++colIt)
      {
         BoundSet& bs(colIt->second);
         for (BoundSet::iterator bsIt(bs.begin());
              bsIt != bs.end();
              ++bsIt)
         {
            const BoundInts& bb(*bsIt);
            DeviceModel::CellSet impliedSet;
            Rf_Cell site(bb.site);
            size_t bottomRow(bb.bottom.GetTileRow());
            size_t topRow(bb.top.GetTileRow());
            string siteType(site.GetCellDef().GetName());
            if (boost::regex_match(siteType, cmtLogicPat)) {
               string group;
               TileSet sourceTiles1;
               if (bb.tile->HasGroups()) {
                  if (bb.tile->GetGroupSet()->size() == 1) {
                     group = "  Group: " + *bb.tile->GetGroupSet()->begin();
                  } else {
                     group = "  Group: {" +
                        SetToString(bb.tile->GetGroupSet()) + "}";
                  }

                  // The I/O FIFO is +/-1.
                  for (size_t adjCol(col - 1);
                       adjCol < col + 2;
                       adjCol += 2)
                  {
                     CmtMap::iterator iofifoColIt(cmtMap.find(adjCol));
                     if (iofifoColIt != cmtMap.end()) {
                        CmtColumnMap& iofifoColMap(iofifoColIt->second);
                        for (CmtColumnMap::iterator
                                iofifoColMapIt(iofifoColMap.begin());
                             iofifoColMapIt != iofifoColMap.end();
                             ++iofifoColMapIt)
                        {
                           size_t fifoBottomRow(iofifoColMapIt->first);

                           if (topRow < fifoBottomRow) {
                              break;
                           }

                           BoundSet& fifoSet(iofifoColMapIt->second);
                           for (BoundSet::iterator fsIt(fifoSet.begin());
                                fsIt != fifoSet.end();
                                ++fsIt)
                           {
                              const BoundInts& fbb(*fsIt);
                              Rf_Cell fsite(fbb.site);
                              size_t fifoTopRow(fbb.top.GetTileRow());

                              if (fifoTopRow < bottomRow) {
                                 break;
                              }

                              if (!group.empty()) {
                                 // The I/O FIFO must be in the same
                                 // group as the CMT site.
                                 if (!fbb.tile->HasGroups() ||
                                     !fbb.tile->Equal(*bb.tile))
                                 {
                                    sourceTiles1.insert(bb.tile);
                                    impliedSet.insert(fsite);
                                    bb.tile->SetHasBadNeighbor();
                                    fbb.tile->SetHasBadNeighbor();
                                 } else {
                                    ; // No violation. Groups match.
                                 }
                              }
                           }
                        }
                     }
                  }
               } else if (bb.tile->HasProgramming()) {
                  // One or more CMT sites is used at top level. This
                  // is a violation.
                  topLevelCmtSet.insert(bb.tile->GetName());
               } else {
                  ; // No violation. An unused CMT site.
               }                              

               if (!impliedSet.empty() && !bb.tile->logicSiteSet.empty()) {
                  const string message(ComposeImpliedViolation(
                                          sourceTiles1,
                                          group,
                                          impliedSet));
                  impliedSiteViolationSet.insert(message);
               }
            }
         }
      }
   }
   topLevelCmtCount = topLevelCmtSet.size();
}

// Check that area range constraints implied by other area range
// constraints are present. This applies to 7 series CMT isolation
// group constraints.
//
// There are only three ways for a horizontal fence to cross the CMT
// column in an I/O bank. See the beautiful picture in the isolation
// analysis documentation.
//
// Two tiles combined contain 10 phaser-related sites. If any one of
// the ten is isolated the other 9 must be isolated the same way. If
// any phaser is used, the PLL above must be constrained as well
// (otherwise implementation will fail). There will be two adjacent
// tiles: CMT_TOP_[RL]_LOWER_T, CMT_TOP_[RL]_UPPER_B.
//
// Isolation constraints on the MMCM and PLL do not imply any phaser
// things need to be constrained (isolated).
//
//    Tile CMT_TOP_R_LOWER_B contains site MMCME2_ADV.
//    Tile CMT_TOP_R_UPPER_T contains site PLLE2_ADV.
//
void TileChecker::CheckImpliedAreaRanges() {
   ASSERT(deviceModel->Is7Series());
   CmtMap cmtMap;
   MakeCmtMap(cmtMap);
   CheckImpliedIOFifos(cmtMap);

   TileSet seedTiles;
   for (size_t i(0); i < grid.size(); ++i) {
      Tile* tile(grid.at(i));
      // These are the site types in the CMT_TOP_[RL]_UPPER_B.
      // Whenever any of these sites appear, several others should
      // also. That set is built up below.
      if (tile && (tile->HasSiteType("PHASER_IN_PHY") ||
                   tile->HasSiteType("PHASER_OUT_PHY") ||
                   tile->HasSiteType("PHASER_REF") ||
                   tile->HasSiteType("PHY_CONTROL")) &&
          !deviceModel->IsInterconnectCell(tile->GetCell()))
      {
         seedTiles.insert(tile);
      }
   }
   // Go through all the sites. If the tile of the site is used,
   // then the site itself must be used, otherwise there will be
   // implementation difficulties. A tile only has a given site type
   // if the site is used (or in UCF mode, assumed to be used) in the
   // design.
   Rf_CellContainer sites(deviceModel->GetSites());
   SiteNameTileMap relatedSites;
   for (Rf_CellContainer::const_iterator sitesIt(sites.begin());
        sitesIt != sites.end();
        ++sitesIt)
   {
      Rf_Cell site(*sitesIt);
      string siteType(site.GetCellDef().GetName());
      if (siteType == "PHASER_IN_PHY" ||
          siteType == "PHASER_OUT_PHY" ||
          siteType == "PHASER_REF" ||
          siteType == "PHY_CONTROL")
      {
         Rf_Cell cell(deviceModel->GetSiteParentTile(site));
         string cellType(cell.GetCellDef().GetName());
         // Get the tile from the cell coordinates.
         Port::UINT32 column(cell.GetTileColumn());
         Port::UINT32 row(cell.GetTileRow());
         Tile* tile(contourMap->At(column, row));
         ASSERT(tile);
         if (seedTiles.find(tile) != seedTiles.end()) {
            string siteName(site.GetName());
            boost::smatch result;
            if (boost::regex_match(siteName, result, phyPat)) {
               size_t y(ToType<size_t>(result.str(3)));
               size_t n(y / 4);
               InsertImpliedPhaserSites(relatedSites, result.str(2), n, tile);
            } else if (boost::regex_match(siteName, result, prpcPat)) {
               size_t n(ToType<size_t>(result.str(3)));
               InsertImpliedPhaserSites(relatedSites, result.str(2), n, tile);
            }
         }
      }
   }
   // For all the phaser sites that are present,
   // check that all the related sites are present.
   TileSet completedTiles;
   // For each actual isolated site... (These come from MapSitesToTiles().)
   for (CohortGroupMapIt cgmIt(isolatedCmtCohort.begin());
        cgmIt != isolatedCmtCohort.end();
        ++cgmIt)
   {
      string group(cgmIt->first);
      SiteNameTileMap& presentMap(cgmIt->second);
      for (SiteNameTileMapIt sntmIt(presentMap.begin());
           sntmIt != presentMap.end();
           ++sntmIt)
      {
         string siteName(sntmIt->first);
         Tile* keyTile(sntmIt->second);
         if (completedTiles.find(keyTile) == completedTiles.end()) {
            StringSet cohort;
            for (SiteNameTileMapIt sntm2It(presentMap.begin());
                 sntm2It != presentMap.end();
                 ++sntm2It)
            {
               string siteName2(sntm2It->first);
               Tile* keyTile2(sntm2It->second);
               if (keyTile2 == keyTile) {
                  cohort.insert(siteName2);
               }
            }
            // Iterate over the related to see if there are any
            // related that weren't in the cohort the user constrained
            // to be isolated.
            StringSet missingSites;
            StringSet presentSites;
            for (SiteNameTileMapIt rIt(relatedSites.begin());
                 rIt != relatedSites.end();
                 ++rIt)
            {
               string siteName3(rIt->first);
               Tile* keyTile3(rIt->second);
               if (keyTile3 == keyTile) {
                  if (cohort.find(siteName3) == cohort.end()) {
                     // We found a site in the set of three associated
                     // CMT tiles that isn't present in the design. In
                     // other words, it is a violation that the
                     // constraint that would have caused this site to be
                     // isolated is missing.
                     missingSites.insert(siteName3);
                  } else {
                     presentSites.insert(siteName3);
                  }
               }
            }
            if (!presentSites.empty() && !missingSites.empty()) {
               string sourceLabel(1 < presentSites.size()
                                  ? "\n    Sites: "
                                  : "\n    Site: ");
               string requireLabel(1 < missingSites.size()
                                   ? "\n    require sites: "
                                   : "\n    requires site: ");
               string msg(WordWrap("\n  Group: " + group +
                                   sourceLabel + ToString(presentSites) +
                                   requireLabel + ToString(missingSites)));
               impliedSiteViolationSet.insert(msg);
            }
            completedTiles.insert(keyTile);
         } else {
         }
      }
   }
   impliedSiteCount = impliedSiteViolationSet.size();
}

// For UCF-mode, check area ranges by checking marked tiles. There are
// three cases: adjacent site ranges lacking a sufficient fence between
// them; overlapping site ranges; and implied site ranges for the use
// of sites cannot be used without other sites without
// implementation-time failures.
//
// These checks are intended to avoid issues at NCD time. Such issues
// would be detected and reported by the implementation tools or by
// IVT in tile adjacency and content checking.
void TileChecker::CheckAreaRanges() {
   // For each pair of tiles forming an adjacency violation
   for (TileFaultIt tfIt(netAdjacencyFaults.begin());
        tfIt != netAdjacencyFaults.end();
        ++tfIt)
   {
      // Pairs of tiles get added to faults if their group sets
      // are not empty and do not match.
      CheckAdjacentAreaRanges(tfIt->first, tfIt->second);
   }

   // For each tile 
   for (size_t i(0); i < grid.size(); ++i) {
      Tile* tile(grid.at(i));
      if (tile) {
         CheckOverlappingAreaRanges(tile);
      }
   }
   if (deviceModel->Is7Series()) {
      CheckImpliedAreaRanges();
   }
   areaRangeFaultCount = rangeViolationSet.size();
}

// For logic columns that cannot be used as a vertical fence, check
// that the tiles straddling the column do not form adjacency
// violations. Must take into account the possibility that two such
// special columns will be adjacent (isDouble).
void TileChecker::SweepSpecialColumn(Rf_Cell site,
                                     StringSet& columnFaultSet,
                                     bool isDouble = false,
                                     size_t excludeMin = static_cast<size_t>(-1),
                                     size_t excludeMax = 0)
{
   Rf_Cell cell(deviceModel->GetSiteParentTile(site));
   Port::UINT32 column(cell.GetTileColumn());
         
   // Check three pairs: across, down right, down left
   // last row, only check pairs across.
   Tile* previousLeft(NULL);
   Tile* previousRight(NULL);
   for (size_t row(0); row < contourMap->RowCount(); ++row) {
      Tile* tile(contourMap->At(column, row));
      if (tile && (row < excludeMin || excludeMax < row)) {
         size_t clockIntColumn(
            contourMap->NextVisibleColumn(column, row, 1));
         if (!IsHiddenTile(clockIntColumn, row)) {
            size_t leftColumn(
               contourMap->NextVisibleColumn(clockIntColumn, row, -1));
            size_t rightColumn(
               contourMap->NextVisibleColumn(clockIntColumn, row, 1));
            if (string(site.GetCellDef().GetName()) == "MTBF2") {
               leftColumn =
                  contourMap->NextVisibleColumn(leftColumn, row, -1);
               rightColumn =
                  contourMap->NextVisibleColumn(rightColumn, row, -1);
            }
            if (isDouble) {
               rightColumn =
                  contourMap->NextVisibleColumn(rightColumn, row, 1);
            }
            Tile* left(contourMap->At(leftColumn, row));
            Tile* right(contourMap->At(rightColumn, row));
            // Check the diagonals
            ASSERT(right);
            if (previousLeft && previousLeft->IsViolation(right)) {
               ASSERT(right);
               previousLeft->SetHasBadNeighbor();
               if (right) {
                  right->SetHasBadNeighbor();
                  columnFaultSet.insert(
                     "    Tile: " + previousLeft->GetName() +
                     " conflicts with " + right->GetName());
               }
            }
            if (previousRight && previousRight->IsViolation(left)) {
               ASSERT(left);
               previousRight->SetHasBadNeighbor();
               if (left) {
                  left->SetHasBadNeighbor();
                  columnFaultSet.insert(
                     "    Tile: " + left->GetName() +
                     " conflicts with " + previousRight->GetName());
               }
            }
            previousLeft = left;
            previousRight = right;
            if (left && left->IsViolation(right)) {
               ASSERT(right);
               left->SetHasBadNeighbor();
               if (right) {
                  right->SetHasBadNeighbor();
                  columnFaultSet.insert(
                     "    Tile: " + left->GetName() +
                     " conflicts with " + right->GetName());
               }
            }
         }
      }
   }
   configColumnFaultCount = columnFaultSet.size();
}

void TileChecker::SweepConfigCenterColumn(Rf_Cell site,
                                          bool isAdjacentToClockBuffers)
{
   if (!sweptConfigCenterColumn) {
      const string siteType(site.GetCellDef().GetName());
      if (boost::regex_match(siteType, deviceModel->configCenterPat)) {
         SweepSpecialColumn(site, configColumnFaultSet,
                            isAdjacentToClockBuffers,
                            configCenterRowMin,
                            configCenterRowMax);
         sweptConfigCenterColumn = true;
      }
   }
}

// The config center interconnect column and the clock buffer
// interconnect column are not exactly supersites, but they
// are treated as such in that neither can be used as a
// fence, so we need a special case to check for tiles that
// would be adjacent but for these special columns.
void TileChecker::SweepClockBufferColumn(Rf_Cell site) {
   if (!sweptClockBufferColumn) {
      const string siteType(site.GetCellDef().GetName());
      if (boost::regex_match(siteType, deviceModel->clockBufferPat)) {
         SweepSpecialColumn(site, clockFaultSet);
         sweptClockBufferColumn = true;
      }
   }
}

// The PCIE tile can be used as a fence. As it turns out, the PCIE_2_1
// is the only supersite wider than one column that can be used as a
// fence. Its interconnect tiles are not individually displayed. If
// any of them has a violation, transfer that to the supersite.
void TileChecker::SweepSuperSites() {
   Rf_CellDef tileDef(deviceModel->GetRfDevice()->GetTopCellDefTile());
   Rf_CellContainer sites(deviceModel->GetSites());

   Rf_Cell clockSite;
   Rf_Cell configCenterSite;

   // For each site
   for (Rf_CellContainer::const_iterator sitesIt(sites.begin());
        sitesIt != sites.end();
        ++sitesIt)
   {
      Rf_Cell site(*sitesIt);
      string siteType(site.GetCellDef().GetName());
      if (deviceModel->IsSpartan6() || deviceModel->Is7Series()) {
         if (IsPrefix("PCIE", siteType.c_str())) {
            Rf_Cell cell(deviceModel->GetSiteParentTile(site));
            ASSERT(cell);
            Tile* tile(diagramEmitter.GetTile(cell));
            ASSERT(tile);

            // Propagate violations from the INTs of the supersite to
            // the supersite.
            DeviceModel::CellSet ints;
            AddCells ac(ints);
            deviceModel->MapIntTiles(ac, site, tileDef);
            for (DeviceModel::CellSetIt iIt(ints.begin());
                 iIt != ints.end();
                 ++iIt)
            {
               Rf_Cell intCell(*iIt);
               Tile* intTile(diagramEmitter.GetTile(intCell));
               if (intTile) {
                  tile->SetGroupSet(
                     groupSetFamily.FindOrMakeUnion(intTile->GetGroupSet(),
                                                    tile->GetGroupSet()));
                  if (intTile->HasBadNeighbor()) {
                     tile->SetHasBadNeighbor();
                  }
               }
            }
            // If the supersite has a bad neighbor, but the supersite
            // does not have an isolation group, record a supersite
            // violation. This supersite cannot be used as a fence.
            if (tile->HasBadNeighbor() && !tile->HasGroups()) {
               const string typeName(site.GetCellDef().GetName());
               const string siteName(site.GetName());
               string violation(
                  "\n  " + typeName + " Site: " + siteName +
                  " has neighbors in multiple isolation groups.");
               for (TileSetIt tsIt(tile->GetNeighbors().begin());
                    tsIt != tile->GetNeighbors().end();
                    ++tsIt)
               {
                  Tile* neighbor(*tsIt);
                  if (neighbor->HasGroups()) {
                     neighbor->SetHasBadNeighbor();
                     string groups(SetToString(neighbor->GetGroupSet()));
                     violation += "\n    Group: " + groups +
                        " Tile: " + neighbor->GetName();
                  }
               }
               superFaultSet.insert(violation);
            }
            // If the site has a bad neighbor, but also has an
            // isolation group, then it will be reported as ordinary
            // adjacency violations.
         }
      }
      // Find the config center column and the clock buffer column.
      if (deviceModel->Is7Series()) {
         // Collect a clock site and config center site for a special
         // column sweep below.
         if (!clockSite &&
             boost::regex_match(siteType, deviceModel->clockBufferPat))
         {
            clockSite = site;
         }
         if (!configCenterSite &&
             boost::regex_match(siteType, deviceModel->configCenterPat))
         {
            configCenterSite = site;
         }
      }
   }
   if (deviceModel->Is7Series()) {
      Rf_Cell ccCell(deviceModel->GetSiteParentTile(configCenterSite));
      Port::UINT32 ccColumn(ccCell.GetTileColumn());
      Port::UINT32 ccRow(ccCell.GetTileRow());

      size_t ccIntColumn(
         contourMap->NextVisibleColumn(ccColumn, ccRow, 1));

      Rf_Cell clockCell(deviceModel->GetSiteParentTile(clockSite));
      Port::UINT32 clockColumn(clockCell.GetTileColumn());
      Port::UINT32 clockRow(clockCell.GetTileRow());

      size_t clockIntColumn(
         contourMap->NextVisibleColumn(clockColumn, clockRow, 1));

      // The delta between ccIntColumn and clockIntColumn can be seen
      // by examining the tile maps.
      bool adjacent(ccIntColumn - clockIntColumn < 7 ||
                    clockIntColumn - ccIntColumn < 7);
      SweepConfigCenterColumn(configCenterSite, adjacent);
      if (!adjacent) {
         SweepClockBufferColumn(clockSite);
      }
   }
   superFaultCount += superFaultSet.size();
}

// If a tile is not isolated from multiple isolation groups and is not
// a valid fence tile, there is a fence violation.
//
// Note: This is a no-op for UCF mode because taint checking requires
// information about actual routing and logic between isolated
// modules, not an assumption based on area range constraints.
void TileChecker::CheckTaints() {
   for (TileVecIt gIt(grid.begin()); gIt != grid.end(); ++gIt) {
      Tile* tile(*gIt);
      if (tile && tile->HasMultipleTaints()) {
         tile->SetHasBadNeighbor();
      }
   }
}

// R#check_irs_separation, R#check_interconnect_routing - Visit all
// pairs of adjacent (non-hidden) tiles. The tile visit procedure
// looks for supposedly isolated nets from two distinct groups
// occupying adjacent tiles, that is to say, not separated by a fence
// tile.
//
// In the worst case, there are two isolation groups composed of one
// tile each at opposite corners of the device, connected by a long
// zig zagging path of invalid fence tiles. If the device is n x m,
// then the longest path connecting the two corners is O(nm). The
// length is approximately m(n - (n - 1)/2) because every other row or
// column needs to be empty except for one tile, otherwise there is a
// shorter path. The approximation arises from the need for a separate
// case for odd and even n and m. Each pass (before the last) advances
// one tile. Although the violation could be detected when the invalid
// fences "meet in the middle", propagation is allowed to continue so
// that Xs can be drawn on every invalid fence tile, not just the
// middle pair.
//
//    R X X X X X X X     R = Red isolation group 
//    : : : : : : : X     B = Black isolation group
//    X X X X X X X X     X = Invalid fence tile
//    X : : : : : : :     : = Valid fence tile
//    X X X X X X X B     n x m = 8 x 5
//
// As a practical matter, we do not expect a design to have many
// invalid fence tiles and we certainly do not expect them to neatly
// snake around an otherwise empty device.
void TileChecker::SweepAllAdjacent() {
   do {
      sweepState = 0;
      contourMap->SweepAllAdjacent();
   } while (sweepState);
   sweepState = 2;
   contourMap->SweepAllAdjacent();
   SweepSuperSites();
   CheckTaints();
}

// Not all PIPs collected in MapOneWireTile() correspond to isolation
// violations. SweepAllAdjacent() propagated isolation group
// information from isolated tiles to invalid fence tiles. If an
// invalid fence tile is associated with more than one group, then it
// is needed as a valid fence, and thus it is an isolation violation.
void TileChecker::CheckTerminals() {
   for (TileVecIt gIt(grid.begin()); gIt != grid.end(); ++gIt) {
      Tile* tile(*gIt);
      if (tile) {
         if (tile->HasIsolatedNet() ||
             tile->HasBusMacro() ||
             tile->HasGroups() ||
             tile->CountTaints() < 2)
         {
            // Used PIPs are acceptable (unavoidable) in isolated
            // tiles (non-fence tiles).
            tile->ClearTerminals();
         }
         irsNetPipCount += tile->HasTerminal();
      }
   }   
}

// R#check_irs_separation, R#check_interconnect_routing
void TileChecker::Check() {
   if (data.NeedConstraintsCheck()) {
      MapSitesToTiles();
   } else {
      MapNetsToTiles();
      CollectNominalFenceTiles();
      MarkComps();
   }
   CollectSuperSiteTileCoordinates();
   AddNeighborsForSuperSites();
   CollectPolygonalBoundaries();
   SweepAllAdjacent();
   CheckSpecialSites();
   if (data.NeedConstraintsCheck()) {
      CheckAreaRanges();
   } else {
      CheckTerminals();
   }
}

void TileChecker::GenerateDiagram() {
   diagramEmitter.GenerateDiagram();
}

// Sort cell names in the naive computer way ignoring embedded numbers.
// For example, INT_X10Y10 < INT_X2Y10.
class CellLess {
public:
   CellLess(TileVec& grid) : grid(grid) {};
   bool operator()(size_t a, size_t b) const {
      if (grid.at(a) && grid.at(b)
          && grid.at(a)->GetCell() && grid.at(b)->GetCell())
      {
         return string(grid.at(a)->GetCell().GetName())
            < string(grid.at(b)->GetCell().GetName());
      } else {
         return a < b;
      }
   };
private:
   TileVec& grid;
};

// R#check_irs_separation, R#check_interconnect_routing - Report tiles
// containing isolation violations. That is to say, tiles containing
// isolated nets from multiple isolation groups.
void TileChecker::ReportMixedTiles() {

   Section("Tiles with Net Content Violations");

   ostringstream violations;
   GroupSetFamily::GroupSetListIt
      beginIt(groupSetFamily.groupSetList.begin());
   GroupSetFamily::GroupSetListIt
      endIt(groupSetFamily.groupSetList.end());
   string sep2;
   for (size_t i(0); i < grid.size(); ++i) {
      Tile* tile(grid.at(i));
      if (tile) {
         Rf_Cell cell(grid.at(i)->GetCell());
         if (tile->HasBadContent() &&
             deviceModel->IsInterconnectCell(cell))
         {
            // We know the cell is good because we checked before we
            // inserted it in the grid.

            // We expect tiles with one isolation group and zero
            // isolation groups (global). Otherwise, there is an
            // isolation failure.
            ++netContentFaultCount;
            Data::GroupSet* gs(tile->GetGroupSet());
            GroupSetFamily::GroupSetListIt gsIt(find(beginIt, endIt, gs));
            ASSERT(gsIt != endIt);
            const string name(cell.GetName());
            const string gsString(" {" + SetToString(gs) + "}");
            const string sep1("\n    ");
            const string
               netsString(SetToString(&tile->netDescriptions, sep1));
            violations <<
               sep2 + 
               "  Tile: " + name
               + gsString + ":" + sep1
               + netsString;
            sep2 = "\n\n";
         }
      }
   }
   if (!violations.str().empty()) {
      Report("Isolation Violation:114: One or more tiles contain\n"
             "nets from multiple isolation groups.");

      if (deviceModel->IsSpartan6()) {
         Report("\n"
                "  Tile names contain coordinate information that does not\n"
                "  directly correspond to site coordinates in all cases. SLICE\n"
                "  coordinates match the TIEOFF blocks visible in FPGA Editor\n"
                "  near the top right of each switch box.");

      }
      Report("\n" + violations.str());
   } else {
      Report("  No tiles containing nets from multiple isolation\n"
             "  groups were found.");
   }
}

// R#check_irs_separation - All inter-region signals must be
// examined. List all signals that are not confined to a single
// isolation region.
//
// Note: the note about the TIEOFF block does not apply to Spartan-6.
void TileChecker::ReportFenceNets() {

   Section("Tiles in the Fence Containing Nets");

   Report("  The tiles below are in the nominal fence as defined by\n"
          "  the region outside all isolation groups. This list is intended \n"
          "  to aid the user in demonstrating that an actual fence has\n"
          "  been realized.\n");

   if (deviceModel->IsSpartan6()) {
      Report("  Tile names contain coordinate information that does not\n"
             "  directly correspond to site coordinates in all cases. SLICE\n"
             "  coordinates match the TIEOFF blocks visible in FPGA Editor\n"
             "  near the top right of each switch box.");
   }

   Report("  Total nets in the nominal fence: " +
          ToString(nominalFenceNets));

   sort(fenceList.begin(), fenceList.end(), CellLess(grid));
   for (IndexVecIt uIt(fenceList.begin());
        uIt != fenceList.end();
        ++uIt)
   {
      Tile* tile(grid.at(*uIt));
      if (tile) {
         Rf_Cell cell(tile->GetCell());
         if (deviceModel->IsInterconnectCell(cell)
             && !tile->netDescriptions.empty())
         {
            Report("\n  Tile: " + string(cell.GetName()) + "\n"
                   + tile->GetNetsBrief());
         }
      }
   }
   if (fenceList.empty()) {
      Report("  All tiles containing nets are in area ranges associated\n"
             "  with isolation groups.");
   }
}

// R#check_irs_separation, R#check_direct_routing, R#check_logic_tiles -
// Report comp blocks in the fence.
void TileChecker::ReportFenceComps() {

   Section("Tiles in the Fence Containing Programming");

   ostringstream violations;
   string sep;
   for (size_t i(0); i < grid.size(); ++i) {
      Tile* tile(grid.at(i));
      if (tile && tile->HasProgramming() &&
          (tile->HasMultipleTaints() || tile->HasViolation()))
      {
         Rf_Cell cell(tile->GetCell());
         if (cell) {
            const string tileName(tile->GetCell().GetName());
            violations <<
               sep + "  Components in tile " + tileName
               + ":\n    Component: "
               + SetToString(&tile->compSet, "\n    Component: ");
            sep = "\n\n";
            logicContentFaultCount += tile->compSet.size();
         } else {
            FatalInternal(124, "No cell for tile " + ToString(i) + ".");
         }
      }
   }
   if (logicContentFaultCount) {
      Report("Isolation Violation:134: The tiles below are used to separate\n"
             "isolated nets or logic and have non-default configurations.");
      Report("\n" + violations.str());
   } else {
      Report("  All the fence tiles are in their default configurations.");
   }
}

// R#check_irs_separation - Report tiles containing wires that have
// PIPs in the fence.
void TileChecker::ReportFencePips() {

   Section("Tiles in the Fence Containing Used PIPs");

   if (!irsNetPipCount) {
      Report("  No wires with terminals in the fence were found.");
   } else {
      Report("Isolation Violation:129: The following tiles are configured to\n"
             "use wires that contain programmable interconnect points (PIPs)\n"
             "in the fence. Such wires are not allowed. Check that\n"
             "constraints needed for isolation are present.");
   }
   for (TileVecIt gIt(grid.begin()); gIt != grid.end(); ++gIt) {
      Tile* tile(*gIt);
      if (tile && tile->HasTerminal()) {
         const string tileName(tile->GetCell().GetName());
         Report("\n  Tile " + tileName + ":");

         // Report the tiles that are not isolated from this one.
         ASSERT(tile->HasTaints());
         TileSet taintSet(tile->GetTaints());
         string influences("    Influences: ");
         string sep;
         for (TileSetIt tsIt(taintSet.begin());
              tsIt != taintSet.end();
              ++tsIt)
         {
            Tile* taint(*tsIt);
            influences += sep + taint->GetName();
            sep = ", ";
         }
         Report(influences);

         NetWireMap nwm(tile->irsPips);
         for (NetWireMapIt nwmIt(nwm.begin()); nwmIt != nwm.end(); ++nwmIt) {
            XCad_CadNetwork* net(nwmIt->first);
            const string netName(net->GetUserHierName());
            const string netGroup(data.netGroup.find(net)->second);
            bool isGlobal(netGroup == "combined" || netGroup == "ignored");
            const string scopeLabel(isGlobal
                                    ? string("Top-level Net: ")
                                    : "Group " + netGroup + " Net: ");
            Report("    " + scopeLabel + netName);
            WireSet ws(nwmIt->second);
            for (WireSetIt wsIt(ws.begin()); wsIt != ws.end(); ++wsIt) {
               Rf_Wire wire(*wsIt);
               Rf_Node node(wire.GetNode());
               const string wireName(wire.GetName());
               const string nodeLabel(string(node.GetName())
                                      + ":" + ToString(node.GetIndex()));
               Report("      Node: " + nodeLabel + " Wire: " + wireName);
            }
         }
      }
   }
}

// R#check_irs_separation, R#check_interconnect_routing - List all
// pairs of tiles that represent distinct non-empty sets of isolation
// groups. Global signals are ignored in this function.
void TileChecker::ReportAdjacentTiles() {

   Section("Tiles with Net Adjacency Violations");

   if (!netAdjacencyFaults.empty()) {
      Report("Isolation Violation:112: "
             "The pairs of tiles listed below contain\n"
             "nets from multiple isolation groups, but lack an adequate\n"
             "fence between them.");
   }
   sort(netAdjacencyFaults.begin(),
        netAdjacencyFaults.end(),
        Tile::TileFaultLess());
   TileFaultIt endIt(unique(netAdjacencyFaults.begin(),
                            netAdjacencyFaults.end(),
                            Tile::TileFaultEqual()));
   netAdjacencyFaultCount =
      static_cast<size_t>(endIt - netAdjacencyFaults.begin());
   if (!netAdjacencyFaults.empty()) {
      for (TileFaultIt tfIt(netAdjacencyFaults.begin());
           tfIt != endIt;
           ++tfIt)
      {
         // Pairs of tiles get added to faults if their group sets
         // are not empty and do not match.
         Tile* t1(tfIt->first);
         Tile* t2(tfIt->second);
         if (ArgParser::verbose) {
            const string t1Name(t1->GetCell().GetName());
            const string t2Name(t2->GetCell().GetName());
            const string t1Set("{" + SetToString(t1->GetGroupSet()) + "}");
            const string t2Set("{" + SetToString(t2->GetGroupSet()) + "}");
            const string t1Nets(SetToString(&t1->netDescriptions, "\n    "));
            const string t2Nets(SetToString(&t2->netDescriptions, "\n    "));
            Report("\n  Tile: " + t1Name + " " + t1Set + ":\n"
                   "    " +  t1Nets
                   + "\n  and tile: " + t2Name + " " + t2Set + ":\n"
                   "    " + t2Nets);
         }
      }
   } else {
      Report("  No tile adjacency violations were found.");
   }
   ConsoleDot();
}

// Report violations related to the 7 series Config Center area. Two
// violations are reported: usage of the config center sites by more
// than one isolation group; and isolated neighbors of the Config
// Center area in more than one isolation group.
void TileChecker::ReportConfigCenterFaults() {
   if (configCenterGroupSet && 1 < configCenterGroupSet->size()) {
      Report("Isolation Violation:143: "
             "The Config Center sites cannot be used by more than one\n"
             "isolation group. The complete list of config center sites is:\n"
             "  BSCAN, CAPTURE, DCIRESET, DNA_PORT, EFUSE_USR,\n"
             "  FRAME_ECC, ICAP, STARTUP, USR_ACCESS, and XADC.\n");
      for (TileSetIt ccsIt(configCenterSet.begin());
           ccsIt != configCenterSet.end();
           ++ccsIt)
      {
         Tile* tile(*ccsIt);
         if (tile->HasGroups()) {
            Report("  Group: " + SetToString(tile->GetGroupSet()) +
                   " tile: " + tile->GetName());
         }
      }
   }
   if ((configCenterNeighborGroupSet &&
        1 < configCenterNeighborGroupSet->size()) ||
       (configCenterNeighborGroupSet && configCenterGroupSet &&
        configCenterNeighborGroupSet != configCenterGroupSet))
   {
      if (configCenterGroupSet && 1 < configCenterGroupSet->size()) {
         Report("");
      }
      Report("Isolation Violation:144: "
             "The Config Center sites cannot be used as a fence.\n"
             "The complete list of config center sites is:\n"
             "  BSCAN, CAPTURE, DCIRESET, DNA_PORT, EFUSE_USR,\n"
             "  FRAME_ECC, ICAP, STARTUP, USR_ACCESS, and XADC.\n");
      for (TileSetIt ccfsIt(configCenterFaultSet.begin());
           ccfsIt != configCenterFaultSet.end();
           ++ccfsIt)
      {
         Tile* tile(*ccfsIt);
         Report("  Group: " + SetToString(tile->GetGroupSet()) +
                " Tile: " + tile->GetName());
      }
   }
}

// R#check_irs_separation, R#check_interconnect_routing - Report
// special case adjacency violations.
void TileChecker::ReportSpecialTiles() {
   if (deviceModel->IsSpartan6() || deviceModel->Is7Series()) {

      Section("Tiles with Special-Case Adjacency Violations");

      string sep;
      if (configCenterFaultCount == 0 &&
          dspFaultCount == 0 &&
          superFaultCount == 0 &&
          gtxFaultCount == 0 && 
          impliedSiteCount == 0 &&
          topLevelCmtCount == 0)
      {
         Report("  No special-case adjacency violations were found.");
      } else {
         if (configCenterFaultCount) {
            ReportConfigCenterFaults();
            sep = "\n";
         }
         if (!s6DspFaultSet.empty()) {
            Report(sep +
                   "Isolation Violation:131: "
                   "Spartan-6 DSP sites from distinct isolation\n"
                   "groups must be vertically separated by "
                   "two unused DSP sites (eight\n"
                   "interconnect block rows).");
            Report("  " + ToString(s6DspFaultSet, "\n    "));
            sep = "\n";
         }
         if (!superFaultSet.empty()) {
            Report(sep +
                   "Isolation Violation:137: "
                   "The following sites cannot be used as a fence.");
            Report("  " + ToString(superFaultSet, "\n    "));
            sep = "\n";
         }
         if (!impliedSiteViolationSet.empty()) {
            Report(sep +
                   "Isolation Violation:141: Isolated CMT"
                   " logic sites require all adjacent\n"
                   "CMT sites to be included in matching isolation"
                   " groups. The following\n"
                   "sites are either not included in isolated area"
                   " ranges or are included\n"
                   "in non-matching isolated ranges.");
            Report("  " + ToString(impliedSiteViolationSet, "\n  "));
            sep = "\n";
         }
         if (!topLevelCmtSet.empty()) {
            Report(sep +
                   "Isolation Violation:142: Used CMT logic sites"
                   " must be isolated.\n"
                   "The following tiles are not part of isolated"
                   " area ranges:");
            Report("  " + ToString(topLevelCmtSet, "\n  "));
            sep = "\n";
         }
         if (!gtxFaultSet.empty()) {
            Report(sep +
                   "Isolation Violation:148: 7 Series GTX tiles"
                   " within a single clock\n"
                   "region cannot occupy multiple isolation groups.\n");
            Report("  " + ToString(gtxFaultSet, "\n  "));
            sep = "\n";
         }
         if (!clockFaultSet.empty()) {
            Report(sep +
                   "Isolation Violation:149: The clock buffer column"
                   " is not allowed\n"
                   "as a fence.\n");
            Report("  " + ToString(clockFaultSet, "\n  "));
            sep = "\n";
         }
         if (!configColumnFaultSet.empty()) {
            Report(sep +
                   "Isolation Violation:150: The config center column"
                   " is not allowed\n"
                   "as a fence.\n");
            Report("  " + ToString(configColumnFaultSet, "\n  "));
            sep = "\n";
         }
      }
   }
}

// R#check_irs_separation, R#check_interconnect_routing
void TileChecker::ReportSummary() {
   string netAdj(       ToString(netAdjacencyFaultCount));
   string logicAdj(     ToString(logicAdjacencyFaultCount));
   string netContent(   ToString(netContentFaultCount));
   string logicContent( ToString(logicContentFaultCount));
   string pips(         ToString(irsNetPipCount));
   string loads(        ToString(data.irsMultiLoadCount));
   string ccs(          ToString(configCenterFaultCount));
   string dsps(         ToString(dspFaultCount));
   string implied(      ToString(impliedSiteCount));
   string topCmt(       ToString(topLevelCmtCount));
   string supers(       ToString(superFaultCount));
   string gtxQuads(     ToString(gtxFaultCount));
   string cbcs(         ToString(clockFaultCount));
   string ccfs(         ToString(configColumnFaultCount));
   Summary("\n  Tile Adjacency\n\n"
           "    Net Adjacency Violations: " +           netAdj + "\n"
           "    Logic Adjacency Violations: " +         logicAdj);
   Summary("\n  Tile Content\n\n"
           "    Net Content Violations: " +             netContent + "\n"
           "    Logic Content Violations: " +           logicContent);
   Summary("\n  Inter-region Signals\n\n"
           "    Inter-region Net PIP Violations: " +    pips + "\n"
           "    Inter-region Load Violations: " +       loads);
   Summary("\n  Special Fence Rules\n\n"
           "    Config Center Violations: " +           ccs + "\n"
           "    DSP Violations: " +                     dsps + "\n"
           "    Implied Site Violations: " +            implied + "\n"
           "    Top-level CMT Violations: " +           topCmt + "\n"
           "    Supersite Violations: " +               supers + "\n"
           "    GTX Violations: " +                     gtxQuads + "\n"
           "    Clock Buffer Column Violations: " +     cbcs + "\n"
           "    Config Center Column Violations: " +    ccfs);
}

// R#check_irs_separation, R#check_interconnect_routing
void TileChecker::ReportFaults() {
   ReportMixedTiles();
   ReportFenceNets();
   ReportFenceComps();
   ReportFencePips();
   ReportAdjacentTiles();
   ReportSpecialTiles();
   ReportSummary();
}

// Used by ConstraintsChecker
void TileChecker::ReportAreaSeparation() {

   Section("Area Group Separation");

   if (data.siteRectMap.empty()) {
      Report("  No area groups found.");
   } else if (rangeViolationSet.empty()) {
      Report("  No area group range violations were found.");
   } else {
      string sep("");
      for (StringSetIt rvsIt(rangeViolationSet.begin());
           rvsIt != rangeViolationSet.end();
           ++rvsIt)
      {
         string rv1(sep + *rvsIt);
         Report(rv1);
         sep = "\n";
      }
   }

   Summary("\n  Area Group Isolation\n\n"
           "    Area Group Isolation Violations: "
           + ToString(areaRangeFaultCount));
}

// Used by ConstraintsChecker. Includes summary.
void TileChecker::ReportSpecialTileFaults() {
   ReportSpecialTiles();
   string ccs(          ToString(configCenterFaultCount));
   string dsps(         ToString(dspFaultCount));
   string implied(      ToString(impliedSiteCount));
   string topCmt(       ToString(topLevelCmtCount));
   string supers(       ToString(superFaultCount));
   string gtxQuads(     ToString(gtxFaultCount));
   string cbcs(         ToString(clockFaultCount));
   string ccfs(         ToString(configColumnFaultCount));
   Summary("\n  Special Fence Rules\n\n"
           "    Config Center Violations: " +           ccs + "\n"
           "    DSP Violations: " +                     dsps + "\n"
           "    Implied Site Violations: " +            implied + "\n"
           "    Top-level CMT Violations: " +           topCmt + "\n"
           "    Supersite Violations: " +               supers + "\n"
           "    GTX Violations: " +                     gtxQuads + "\n"
           "    Clock Buffer Column Violations: " +     cbcs + "\n"
           "    Config Center Column Violations: " +    ccfs);
}

int TileChecker::GetAreaRangeFaultCount() {
   return areaRangeFaultCount;
}

int TileChecker::GetErrorCount() {
   return 
      areaRangeFaultCount +
      configCenterFaultCount +
      dspFaultCount +
      impliedSiteCount +
      irsMultiLoadCount +
      irsNetPipCount +
      logicAdjacencyFaultCount +
      logicContentFaultCount +
      netAdjacencyFaultCount +
      netContentFaultCount +
      superFaultCount +
      topLevelCmtCount +
      gtxFaultCount +
      clockFaultCount +
      configColumnFaultCount;
}

// A la Unix, return 0 for success and 1 for failures.
int TileChecker::GetStatus() {
   return !!GetErrorCount();
}
